"""This is a raw python version of mandel.xml

The purpose of this code is to provide a speed comparison against moya code.

"""
from __future__ import print_function
from __future__ import division

import sys

PY2 = sys.version_info[0] == 2

if not PY2:
    xrange = range


def mandel(xsize=80, ysize=20, max_iteration=50):
    chars = " .,~:;+*%@##"
    for pixy in xrange(ysize):
        y0 = (pixy / ysize) * 2 - 1
        row = ""
        for pixx in xrange(xsize):
            x0 = (pixx / xsize) * 3 - 2
            x = 0
            y = 0
            iteration = 0
            while (x * x + y * y < 4) and iteration < max_iteration:
                xtemp = x * x - y * y + x0
                y = 2 * x * y + y0
                x = xtemp
                iteration += 1
            row += chars[iteration % 10]
        print(row)

if __name__ == "__main__":
    from moya.tools import timer
    with timer("Calculate mandelbrot set in raw python"):
        mandel()




"""Simple text based progress bar"""


from __future__ import unicode_literals
from __future__ import division


from time import sleep


class Progress(object):
    """Renders a progress bar to the console"""
    def __init__(self, console, msg, num_steps=100, width=12, indent='', vanish=False):
        self.console = console
        self.msg = msg
        self.num_steps = num_steps
        self.max_line_size = 0
        self.complete = 0.0
        self._step = 0
        self.width = width
        self.indent = indent
        self.vanish = vanish

    def set_num_steps(self, num_steps):
        self.num_steps = num_steps

    def step(self, count=1, msg=None):
        self.update(self._step + count, msg)

    def __call__(self, iterable):
        for item in iterable:
            yield item
            self.step()

    def update(self, step, msg=None):
        if step is not None:
            self._step = step
        if self.num_steps is None:
            complete = 0.0
        elif not self.num_steps:
            complete = 1.0
        else:
            complete = float(self._step) / self.num_steps
        self.complete = min(complete, 1.0)
        if msg is not None:
            self.msg = msg
        self.render()
        # if self.step == self.num_steps:
        #     self.done()
        # else:
        #     self.render()

    def render(self, line_end='\r'):
        if not self.console.is_terminal():
            return
        progress = "{}%".format(int(self.complete * 100.0)).ljust(4)
        num_bars = int(self.complete * self.width)
        bars = '=' * num_bars
        bars = bars.ljust(self.width, ' ')
        out = "{indent}{progress} [{bars}]".format(indent=self.indent,
                                                   bars=bars,
                                                   progress=progress)
        if self.msg:
            out = "\r{1} {0}".format(self.msg, out)

        out = out.expandtabs()

        self.max_line_size = max(len(out), self.max_line_size)
        out = out.ljust(self.max_line_size + 1, ' ') + line_end

        self.console(out).flush()

    def done(self, msg=None):
        if msg is not None:
            self.msg = msg
        if self.vanish:
            self.render()
            sleep(0.1)
            self.console(len(self.msg) * ' ')('\r')
        else:
            if not self.console.is_terminal():
                self.console.text(self.msg)
            else:
                self.render(line_end='\n')


class ProgressContext(object):
    def __init__(self, progress):
        self.progress = progress

    def __enter__(self):
        self.progress.console.show_cursor(False)
        self.progress.render()
        return self.progress

    def __exit__(self, *args, **kwargs):
        self.progress.done()
        self.progress.console.show_cursor(True)


if __name__ == "__main__":
    from time import sleep
    from moya.console import Console
    c = Console()
    p = Progress(c, "Extracting...", 100)
    p.render()

    for step in xrange(100):
        sleep(.01)
        p.step()
    p.render()
    print



"""A system to create a directory tree from a template.

The format is very simple. A line beginning with @ should be a path to a file. Subsequent lines after that line will be appended to the specified file.

The template is passed through moya templates, creating a very flexible system for dynamically creating file trees.

In order to allow these fs templates to generate moya templates, the syntax is slightly different.
The filesystem template syntax should be {{% %}} for logic, and ${{ }} for substitution.


"""


from __future__ import unicode_literals
from __future__ import print_function

from ...context import Context
from ...template.moyatemplates import Template

from fs.path import dirname, join, relpath

import re


def compile_fs_template(fs, template_text, data=None, path=None):
    """Compile a fs template structure in to a filesystem object"""
    if data is None:
        data = {}
    template = Template(template_text)
    template.re_special = re.compile(r'\{\{\%((?:\".*?\"|\'.*?\'|.|\s)*?)\%\}\}|(\{\{\#)|(\#\}\})')
    context = Context(re_sub=r'\$\{\{(.*?)\}\}')
    #with context.frame("data"):
    fs_template = template.render(data, context=context)

    out_type = None
    out_filename = None
    file_lines = []

    def write_file(filename, file_type):
        if filename:
            if file_type.lower() == "text":
                with fs.open(filename, 'wt') as f:
                    f.write('\n'.join(file_lines) + '\n')
            elif file_type.lower() == "wraptext":
                import textwrap
                with fs.open(filename, 'wt') as f:
                    for line in file_lines:
                        f.write('\n'.join(textwrap.wrap(line, 79)) + '\n')
            elif file_type.lower() == "bin":
                with fs.open(filename, 'wb') as f:
                    for line in file_lines:
                        chunk = b''.join(chr(int(a + b, 16)) for a, b in zip(line[::2], line[1::2]))
                        f.write(chunk)

            del file_lines[:]

    for line in fs_template.splitlines():
        line = line.rstrip()
        if line.startswith('@'):
            write_file(out_filename, out_type)
            out_filename = None
            out_type, path_spec = line[1:].split(' ', 1)
            if path:
                path_spec = join(path, relpath(path_spec))
            if path_spec.endswith('/'):
                fs.makedir(path_spec, allow_recreate=True, recursive=True)
                out_filename = None
            else:
                fs.makedir(dirname(path_spec), allow_recreate=True, recursive=True)
                out_filename = path_spec
            continue
        if out_filename:
            file_lines.append(line)
    if out_filename:
        write_file(out_filename, out_type)


if __name__ == "__main__":
    template = """
@test.txt
This
is a test file
{{%- if readme %}}
@readme.txt
Readme file
-----------
${{message}}
{{%- endif %}}
@templates/base.html
<h1>${title}</h1>
<ul>
    {% for fruit in fruits %}
    <li>${fruit}</li>
    {% endfor %}
</ul>
@settings/production.ini
@foo/bar/baz/
@author
Bob
    """

    from fs.osfs import OSFS
    from fs.memoryfs import MemoryFS

    fs = OSFS('__test__', create=True)
    fs = MemoryFS()
    td = dict(message="Hello, World!", readme=True)
    compile_fs_template(fs, template, td)

    fs.tree()



from __future__ import unicode_literals
from __future__ import print_function

from ..elements.elementbase import LogicElement, Attribute
from ..tags.context import DataSetter
from ..interface import AttributeExposer
from ..compat import iteritems, text_type


class Feed(AttributeExposer):

    __moya_exposed_attributes__ = ['format', 'title', '_items', 'channel', 'xml']

    def __init__(self, format, url, title, description, link, **kwargs):
        self.format = format
        self.url = url
        self.channel = kwargs
        self.channel.update(title=title, description=description, link=link)
        self._items = []
        super(Feed, self).__init__()

    def __repr__(self):
        return "<feed-{} '{}'>".format(self.format, self.title)

    @property
    def title(self):
        return self.channel.get('title', None)

    def add_item(self, item):
        self._items.append(item)

    @property
    def xml(self):
        return self.to_xml()

    def to_xml(self):
        from lxml import etree as ET
        feed = ET.Element('rss', version="2.0", nsmap={'atom': 'http://www.w3.org/2005/Atom'})
        channel = ET.SubElement(feed, 'channel')
        ET.SubElement(channel, "{http://www.w3.org/2005/Atom}link", href=self.url, type="application/rss+xml", rel="self")
        for k, v in iteritems(self.channel):
            if v:
                node = ET.SubElement(channel, k)
                node.text = text_type(v)

        for item in self._items:
            item_node = ET.SubElement(channel, 'item')
            for k, v in iteritems(item):
                if v is not None:
                    node = ET.SubElement(item_node, k)
                    node.text = v
        xml = ET.tostring(feed, encoding='utf-8')
        return xml

    def __xml__(self):
        return self.to_xml()


# http://validator.w3.org/feed/docs/rss2.html
class FeedElement(DataSetter):
    """
    Create a [url http://en.wikipedia.org/wiki/RSS]RSS[/url] Feed object.

    You can add items to a feed with [tag]add-feed-item[/tag].

    """

    class Help:
        synopsis = "create an RSS feed"
        example = """
        <view libname="view.feed">
            <!-- create feed object -->
            <get-fq-url name="list" dst="blog_url" />
            <feed title="${.app.settings.title or 'blog'}"
                description="${.app.settings.description or 'blog feed'}"
                link="${blog_url}" dst="feed"/>

            <db:query model="#Post" dst="posts" orderby="-published_date"
                filter="#Post.published==True" maxresults="25"/>

            <!-- add items -->
            <for src="posts" dst="post">
                <get-fq-url name="showpost" let:slug="post.slug" dst="post_url"/>
                <process-markup src="post.content" type="${.app.settings.default_markup}" dst="description"/>
                <add-feed-item src="feed"
                    title="post.title" link="post_url" description="description"
                    pub_date="post.published_date"/>
            </for>

            <!-- serve feed XML -->
            <serve-xml content_type="application/rss+xml" obj="feed"/>
        </view>
        """

    class Meta:
        tag_name = "feed"

    dst = Attribute("Destination object", type="reference", default=None)

    title = Attribute("Title of Feed", required=True)
    description = Attribute("Description of feed", required=True)
    link = Attribute("Link", required=True, default=None)
    language = Attribute("Language", required=False, default=None)

    def logic(self, context):
        params = self.get_parameters(context)
        optional = {}
        if params.language is None:
            lang = context['.locale.language']
            if lang:
                optional['language'] = text_type(lang)
        optional = self.get_let_map(context)
        if 'generator' not in optional:
            optional['generator'] = text_type(context['.app.lib.version_name'])
        feed = Feed(format='rss',
                    url=context['.request.url'],
                    title=params.title,
                    description=params.description,
                    link=params.link,
                    **optional)
        self.set_context(context, params.dst, feed)


class AddFeedItem(LogicElement):
    """
    Add an item to a [tag]feed[/tag].
    """

    class Help:
        synopsis = "add an item to an RSS feed"

    class Meta:
        one_of = [('title', 'description')]

    src = Attribute("feed to add to", type="expression", required=True)
    title = Attribute("Title of Feed", type="expression", required=False)
    link = Attribute("Link to this item on the web", type="expression", required=True, default=None)
    description = Attribute("Item description", type="expression", required=False)
    author = Attribute("Email address of the author of this item", type="expression", required=False, default=None)
    category = Attribute("Includes the item in one or more categories", type="expression", required=False, default=None)
    guid = Attribute("A string that uniquely identifies the item.", type="expression", required=False, default=None)
    pub_date = Attribute("Indicates when the item was published", type="expression", required=False, default=None)

    _item_names = [('title', None),
                   ('link', None),
                   ('description', None),
                   ('author', None),
                   ('category', None),
                   ('guid', None),
                   ('pub_date', 'pubDate')]

    def logic(self, context):
        params = self.get_parameters(context)
        item = {}
        for name, item_name in self._item_names:
            if self.has_parameter(name):
                if item_name is None:
                    item_name = name
                item[item_name] = params[name]

        if 'pubDate' in item:
            dt = item['pubDate']
            if hasattr(dt, 'rfc2822'):
                item['pubDate'] = dt.rfc2822

        if 'guid' not in item:
            guid = item.get('link', None)
            item['guid'] = guid

        feed = params.src
        try:
            feed.add_item(item)
        except Exception as e:
            self.throw('add-feed-item.add-fail', 'unable to add feed item ({})'.format(e))



"""
Python 2/3 compatibility layer

http://lucumr.pocoo.org/2013/5/21/porting-to-python-3-redux/

"""

import sys

PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] == 3

if PY2:
    text_type = unicode
    binary_type = str
    string_types = (basestring,)
    xrange = xrange
    unichr = unichr
    int_types = (int, long)
    py2bytes = lambda s: s.encode('utf-8')
    raw_input = raw_input
else:
    text_type = str
    binary_type = bytes
    xrange = range
    string_types = (str, bytes)
    unichr = chr
    int_types = (int,)
    py2bytes = lambda s: s
    raw_input = input

if PY2:
    iterkeys = lambda d: d.iterkeys()
    itervalues = lambda d: d.itervalues()
    iteritems = lambda d: d.iteritems()
else:
    iterkeys = lambda d: iter(d.keys())
    itervalues = lambda d: iter(d.values())
    iteritems = lambda d: iter(d.items())


if PY2:
    import urlparse
    from urlparse import urlparse, parse_qs, urlunparse
    from urllib import urlencode, quote, quote_plus
    from itertools import izip_longest as zip_longest
    #import cookie
    from urllib import urlopen
    import SocketServer as socketserver
else:
    from urllib.parse import urlparse, parse_qs, urlunparse
    from urllib.parse import urlencode, quote, quote_plus
    from itertools import zip_longest
    #from http import cookies as cookie
    from urllib.request import urlopen
    import socketserver


if PY2:
    import cPickle as pickle
else:
    import pickle

if PY2:
    from imp import reload
else:
    try:
        from importlib import reload
    except ImportError:
        from imp import reload

if PY2:
    def implements_to_string(cls):
        cls.__unicode__ = cls.__str__
        cls.__str__ = lambda x: x.__unicode__().encode('utf-8')
        return cls
else:
    implements_to_string = lambda x: x


if PY2:
    def implements_iterator(cls):
        cls.next = cls.__next__
        del cls.__next__
        return cls
else:
    implements_iterator = lambda x: x

if PY2:
    def implements_bool(cls):
        cls.__nonzero__ = cls.__bool__
        del cls.__bool__
        return cls
else:
    def implements_bool(cls):
        return cls


if PY2:
    number_types = (int, long, float)
else:
    number_types = (int, float)

if PY2:
    next_method_name = "next"
else:
    next_method_name = "__next__"


if PY2:
    cmp = cmp
else:
    def cmp(a, b):
        if a < b:
            return -1
        elif a == b:
            return 0
        else:
            return +1


class string(object):
    lowercase = 'abcdefghijklmnopqrstuvwxyz'
    uppercase = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    digits = '0123456789'
    punctuation = '!"#$%&\'()*+,-./:;<=>?@[\\]^_`{|}~'


def with_metaclass(meta, *bases):
    class metaclass(meta):
        __call__ = type.__call__
        __init__ = type.__init__

        def __new__(cls, name, this_bases, d):
            if this_bases is None:
                return type.__new__(cls, name, (), d)
            return meta(name, bases, d)
    return metaclass('temporary_class', None, {})



from __future__ import unicode_literals
from __future__ import print_function

from moya.elements.elementbase import LogicElement, Attribute
from moya.tags.context import ContextElementBase, DataSetter
from moya import logic
from moya.console import Cell
from moya.containers import OrderedDict
from moya.template.rendercontainer import RenderContainer
from moya.tools import make_id, url_join
from moya import namespaces
from moya.content import push_content, Content, Node
from moya.render import HTML
from moya.request import UploadFileProxy
from moya.compat import iteritems, itervalues, implements_to_string, string_types, text_type
from moya.context.tools import to_expression
from moya import errors
from moya import http
from moya.interface import AttributeExposer

from collections import defaultdict
from itertools import groupby
from operator import attrgetter
from cgi import FieldStorage
import hashlib

import logging
security_log = logging.getLogger('moya.security')


class SelectOption(object):
    def __init__(self, value, text, group=None, help=None, renderable=None):
        self.value = value
        self.text = text
        self.group = group
        self.help = help
        self.renderable = renderable

    def __repr__(self):
        return "<option {} '{}'>".format(self.value, self.text)

    def __iter__(self):
        return iter([self.value, self.text, self.group])


class Field(Node):
    moya_render_targets = ["html"]

    @classmethod
    def _null_process(cls, context, value=None, **kwargs):
        return value

    def __init__(self,
                 app,
                 label,
                 name,
                 fieldname,
                 src,
                 dst,
                 initial,
                 required,
                 visible,
                 hidelabel=False,
                 default=None,
                 multiple=False,
                 template=None,
                 process_value=None,
                 adapt_value=None,
                 style="paragraphs",
                 data=None,
                 **params):
        self.app = app
        self.label = label
        self.hidelabel = hidelabel
        self.name = name
        self.fieldname = fieldname
        self.src = src
        self.dst = dst
        self.initial = initial
        self.required = required
        self.visible = visible
        self.default = default
        self.multiple = multiple
        self.errors = []
        self._value = None
        self._values = None
        self.id = None
        self.current_group = None
        self.process_value = process_value or self._null_process
        self.adapt_value = adapt_value or self._null_process
        self.style = style
        self.template = template
        self.data = data
        self._choices = []
        self.__dict__.update(params)
        self.is_form_field = True
        super(Field, self).__init__(name)

    def __repr__(self):
        return "<field %s '%s'>" % (self.fieldname, self.name)

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value
        return self

    @property
    def values(self):
        """Gets the value as a list"""
        # Will massage non lists in to a list of one, and a None value as an empty list
        value = self.value
        if value is None:
            return []
        if isinstance(value, (list, tuple)):
            return value
        else:
            return [value]

    def get_default(self):
        return None

    def get_context_value(self, context):
        value = context.get(self.name, None)
        if isinstance(value, list) and not self.multiple:
            if value:
                value = value[0]
            else:
                value = None
        return value

    def process_context_value(self, context):
        if self.name not in context:
            return self.default
        value = context.get(self.name, None)
        if value is None:
            return
        if isinstance(value, list) and not self.multiple:
            if value:
                value = value[0]
            else:
                value = None
        return value

    def add_option(self, value, text, group=None, help=None, renderable=None):
        self._choices.append(SelectOption(value, text, group or self.current_group, help, renderable))

    def set_group(self, label):
        self.current_group = label

    @property
    def choices(self):
        return self._choices

    @property
    def groups(self):
        for key, iter_choices in groupby(self._choices, key=attrgetter('group')):
            yield key, list(iter_choices)

    def set_template_base(self, template_base):
        if not self.template:
            self.template = url_join(template_base, 'fields', '%s.html' % self.fieldname)

    @classmethod
    def get_template_path(cls, style, fieldname):
        return url_join("/moya.forms/styles", style, "fields", "%s.html" % fieldname)

    def moya_render(self, archive, context, target, options):
        template = [options.get('template', None)]
        style = options.get('style', None) or self.style
        template.append(self.get_template_path(style, self.fieldname))
        if style is not 'simple':
            template.append(self.get_template_path('simple', self.fieldname))

        r = RenderContainer.create(archive.find_app('moya.forms'),
                                   template=template)
        r['self'] = self
        r['field'] = self
        if self.data is not None:
            r.update(self.data)
        return r.moya_render(archive, context, target, options)


class ErrorContainer(defaultdict):
    def __init__(self, fields):
        self._fields = fields
        super(ErrorContainer, self).__init__(list)

    def __setitem__(self, k, v):
        if isinstance(v, string_types):
            super(ErrorContainer, self).__setitem__(k, [v])
            self._fields[k][0].errors[:] = [v]
        else:
            super(ErrorContainer, self).__setitem__(k, v)
            self._fields[k][0].errors.append(v)


class FormRenderable(object):
    def __init__(self, renderable=None):
        self.renderable = renderable
        self.children = []
        self.parent = None

    def __repr__(self):
        return "<render %s>" % self.renderable

    def add_child(self, renderable):
        node = FormRenderable(renderable)
        self.children.append(node)
        node.parent = self

    def moya_render(self, archive, context, target, options):
        return self.renderable.moya_render(archive, context, target, options)


class RootFormRenderable(FormRenderable):

    def __init__(self, form, template):
        self.form = form
        self.template = template
        super(RootFormRenderable, self).__init__()

    def moya_render(self, archive, context, target, options):
        td = {'form': self.form,
              'self': self}
        engine = archive.get_template_engine('moya')
        rendered = HTML(engine.render(self.template, td, base_context=context))
        return rendered


class Form(AttributeExposer):
    moya_render_targets = ['html']

    __moya_exposed_attributes__ = sorted(['action',
                                          'app',
                                          'bound',
                                          'content',
                                          'id',
                                          'template',
                                          'csrf_check',
                                          'class',
                                          'method',
                                          'csrf',
                                          'csrf_token',
                                          'data',
                                          'element',
                                          'enctype',
                                          'raw_data',
                                          'error',
                                          'errors',
                                          'fail',
                                          'fields',
                                          'legend',
                                          'ok',
                                          'style',
                                          'validated',
                                          'html_id'])

    def __init__(self, element, context, app, style, template, action, method, enctype, csrf=True, _class=None, html_id=None, legend=None):
        super(Form, self).__init__()
        self.element = element
        self.context = context
        self.app = app
        self.style = style
        self.template = template
        self.action = action
        self.method = method
        self.enctype = enctype
        self.csrf = csrf
        setattr(self, 'class', _class)
        self.html_id = html_id

        self._fields = OrderedDict()
        self.bound = False
        self.validated = False
        self.raw_data = {}
        self._data = {}
        self.error = None
        self.errors = ErrorContainer(self._fields)
        self.id = make_id()
        self.current_field_id = 1

        self.legend = element.legend(context) if legend is None else legend

        self.root = RootFormRenderable(self, template)
        self.current_node = self.root
        self.content = Content(app, template)
        self.content.new_section('fields', None)

        self.field_validators = defaultdict(list)
        self.field_adapters = defaultdict(list)
        self.field_applyers = defaultdict(list)

    def add_csrf(self):
        context = self.context
        field_data = {
            'name': '_moya_csrf',
            'fieldname': 'hidden',
            'src': None,
            'dst': None,
            'initial': self.csrf_token,
            'required': True,
            'label': 'csrf',
            'visible': False,
            'type': 'text',
            '_if': True
        }
        field = self.add_field(field_data, template=None)
        content = context['.content']
        context['field'] = field
        content.add_renderable('hiddeninput', field)

    def __repr__(self):
        return "<form '{}'>".format(self.element.libid)

    def reset(self):
        """Reset the form to a blank state"""
        self._data = {}
        self.raw_data = {}
        for field in self.all_fields:
            field.value = None

    @property
    def csrf_token(self):
        """Return a csrf token"""
        context = self.context
        user_id = text_type(context['.session_key'] or '')
        form_id = self.element.libid
        secret = text_type(self.element.archive.secret)
        raw_token = "{}{}{}".format(user_id, secret, form_id).encode('utf-8', 'ignore')
        m = hashlib.md5()
        m.update(raw_token)
        token_hash = m.hexdigest()
        return token_hash

    def validate_csrf(self, context):
        """Validate CSRF token and raise forbidden error if it fails"""
        if not self.csrf:
            return
        if context['.user'] and context['.request.method'] in ('POST', 'PUT', 'DELETE'):
            csrf = self.raw_data['_moya_csrf']
            #csrf = context['.request.POST._moya_csrf']
            if csrf != self.csrf_token:
                request = context['.request']
                if request:
                    security_log.info('''CSRF detected on request "%s %s" referer='%s' user='%s\'''',
                                      request.method, request.url, request.referer, context['.user.username'])
                raise logic.EndLogic(http.RespondForbidden())

    @property
    def csrf_check(self):
        if not self.csrf:
            return True
        context = self.context
        if context['.user'] and context['.request.method'] in ('POST', 'PUT', 'DELETE'):
            csrf = self.raw_data['_moya_csrf']
            #csrf = context['.request.POST._moya_csrf']
            if csrf != self.csrf_token:
                request = context['.request']
                if request:
                    security_log.info('''CSRF detected on request "%s %s" referer='%s' user='%s\'''',
                                      request.method, request.url, request.referer, context['.user.username'])
                return False
        return True

    @property
    def renderables(self):
        return self.root.children

    @property
    def ok(self):
        return bool(self.validated and not self.errors and not self.error)

    @property
    def fail(self):
        return bool(self.errors or self.error)

    @property
    def fields(self):
        return [field_list[0] for field_list in itervalues(self._fields)]

    @property
    def fields_map(self):
        return {k: v[0] for k, v in self._fields.items()}

    def get_value(self):
        return self

    @property
    def data(self):
        return self._data

    def get_data(self, name):
        value = self.data[name]
        if isinstance(value, FieldStorage):
            value = UploadFileProxy(value)
        return value

    @property
    def current_section(self):
        return self

    def on_content_insert(self, content):
        content.merge(self.content, ignore_sections=["fields"])

    def update_field_value(self, name, value):
        for field in self._fields[name]:
            field.value = value

    def set_field_data(self, name, value):
        self._data[name] = value

    def update(self, value_map):
        if value_map:
            for k, v in value_map.items():
                if k in self._fields:
                    self.update_field_value(k, v)
    __moyaupdate__ = update

    def __contains__(self, key):
        return key in self._fields and self._fields[key] is not None

    @property
    def all_fields(self):
        for field_list in itervalues(self._fields):
            for field in field_list:
                yield field

    def add_field(self,
                  params,
                  enctype=None,
                  style=None,
                  template=None,
                  default=None,
                  process_value=None,
                  adapt_value=None,
                  data=None):
        if enctype is not None and self.enctype is None:
            self.enctype = enctype
        style = style or params.pop('style', None) or self.style
        field = Field(self.app,
                      default=default,
                      template=template,
                      process_value=process_value,
                      adapt_value=adapt_value,
                      style=style,
                      data=data,
                      **params)

        field.id = "field{}_{}".format(self.id, self.current_field_id)
        self.current_field_id += 1
        if params.get('name'):
            name = params['name']
            self._fields.setdefault(name, []).append(field)
        return field

    def add_field_validator(self, field_name, element):
        self.field_validators[field_name].append(element)

    def add_field_adapter(self, field_name, element):
        self.field_adapters[field_name].append(element)

    def add_field_applyer(self, field_name, element):
        self.field_applyers[field_name].append(element)

    def get_field_validators(self, field_name):
        return self.field_validators[field_name]

    def get_field_adapters(self, field_name):
        return self.field_adapters[field_name]

    def push_node(self):
        self.current_node = self.current_node.children[-1]

    def pop_node(self):
        self.current_node = self.current_node.parent

    def add_renderable(self, name, renderable):
        self.current_node.add_child(renderable)

    def add_fail(self, field, msg):
        """Add a message to a list of errors for a given field"""
        self._fields[field][0].errors.append(msg)
        self.errors[field].append(msg)

    def set_fail(self, field, msg):
        """Replace the current list of errors for a field"""
        self._fields[field][0].errors[:] = [msg]
        self.errors[field][:] = [msg]

    def __moyaconsole__(self, console):
        if self.bound:
            console.text("Bound form", fg="cyan", bold=True)
        else:
            console.text("Unbound form", fg="blue", bold=True)

        table = []
        for field in self.all_fields:
            if field.value is None:
                v = Cell('None', dim=True)
            else:
                v = field.value
            table.append([field.name, v])
        console.table(table, ['name', 'value'])

        if self.validated:
            if self.errors:
                console.text("Form errors", fg="red", bold=True)
                error_table = [(field, '\n' .join('* %s' % e.strip() for e in _errors))
                               for field, _errors in iteritems(self.errors)]
                console.table(error_table, ["field", "error"])
            else:
                if self.error:
                    console.text('Form error "%s"' % self.error.strip(), fg="red", bold=True)
                else:
                    console.text("Validated, no errors", fg="green", bold=True)

        if not self.csrf_check:
            console.text('CSRF check failed -- form did not originate from here!', fg="red", bold=True)

    # def fill_initial(self, context):
    #     for field in self.all_fields:
    #         if field.initial is not None:
    #             value = field.initial
    #             field.value = self.raw_data[field.value] = value

    def get_initial_binding(self, context):
        binding = {}
        for field in self.all_fields:
            if field.initial is not None:
                value = field.initial
                binding[field.name] = field.adapt_value(context, value=value)
        return binding

    def fill(self, obj):
        for field in self:
            field.value = obj.get(field.name, None)

    def get_binding(self, context, bind):
        binding = {}
        if not bind:
            return binding

        with context.data_frame(bind):
            for field in self.all_fields:
                if field.name:
                    binding[field.name] = field.process_context_value(context)
        return binding

    def get_src_binding(self, context, bind_root):
        binding = {}
        with context.frame(bind_root):
            for field in self.all_fields:
                if field.src and field.src in context:
                    value = context[field.src]
                    value = field.process_value(context, value=value)
                    binding[field.name] = value
        return binding

    def bind(self, context, *bindings):
        self.set_data(context, *bindings)
        self.bound = True

    def set_data(self, context, *data):
        for field in self.all_fields:
            for binding in reversed(data):
                if field.name in binding:
                    value = binding[field.name]
                    if value is not None:
                        field.value = self.raw_data[field.name] = value
                        break

    def moya_render(self, archive, context, target, options):
        form_template = self.template
        if form_template is None:
            form_template = "/moya.forms/styles/%s/form.html" % self.style
        self.content.template = form_template
        self.content.td['form'] = self
        return self.content.moya_render(archive, context, target, options)


class Reset(ContextElementBase):
    """
    Resets a form to a blank state.

    """
    xmlns = namespaces.forms

    class Help:
        synopsis = "reset a form"

    src = Attribute("Form source", type="index", default="form", evaldefault=True, map_to="src")

    def logic(self, context):
        form = self.src(context)
        form.reset()


class Bind(ContextElementBase):
    """
    Bind a form to data

    """
    xmlns = namespaces.forms

    class Help:
        synopsis = "Add data to a form"

    bind = Attribute("Object to bind to", type="expression", required=False)
    src = Attribute("Source object to fill in fields", type="expression", default=None)

    def logic(self, context):
        bind, src = self.get_parameters(context, 'bind', 'src')
        if bind is None:
            bind = {}
        bind.update(self.get_let_map(context))
        form = src
        form.bind(context, bind)


class Get(DataSetter):
    """
    Get a form object.

    """
    xmlns = namespaces.forms

    class Help:
        synopsis = "get a form"
        example = """
        <forms:get form="#form.login" dst="form" />
        """

    bind = Attribute("Object to bind to", type="expression", default=".request.method == 'POST' ? .request.multi.POST : None", evaldefault=True)
    form = Attribute("Form reference", required=True)
    src = Attribute("Source object to fill in fields", type="reference", default=None)
    dst = Attribute("Destination to store form", required=False, default=None)
    style = Attribute("Override form style", required=False, default=None)
    template = Attribute("Override form template", required=False, default=None)
    action = Attribute("Form action", required=False, default=None)
    method = Attribute("Form method", required=False, default="post", choices=['get', 'post'])
    withscope = Attribute("Use current scope?", default=False, type="boolean")
    blank = Attribute("Blank form?", default=False, type="boolean")
    _id = Attribute("Override HTML id attribute", type="text", default=None)
    legend = Attribute("Override form Legend", type="text", default=None)

    def logic(self, context):
        (bind,
         form,
         src,
         dst,
         style,
         template,
         action,
         method,
         withscope,
         blank,
         _id,
         legend) = self.get_parameters(context,
                                       'bind',
                                       'form',
                                       'src',
                                       'dst',
                                       'style',
                                       'template',
                                       'action',
                                       'method',
                                       'withscope',
                                       'blank',
                                       'id',
                                       'legend')
        if withscope:
            scope = context['.call']
        call = self.push_funccall(context)
        try:
            yield logic.DeferNodeContents(self)
            call.update(self.get_let_map(context))

            args, kwargs = call.get_call_params()
            if withscope:
                new_kwargs = scope
                new_kwargs.update(kwargs)
                kwargs = new_kwargs

        finally:
            self.pop_funccall(context)

        app = self.get_app(context)
        form_app, form_element = self.get_element(form, app)
        app = form_app or app

        template = app.resolve_template(template)

        with self.call(context, app, **kwargs) as call:
            (form_style,
             form_template,
             form_action,
             form_method,
             _class,
             enctype,
             csrf,
             form_id) = form_element.get_parameters(context,
                                                    'style',
                                                    'template',
                                                    'action',
                                                    'method',
                                                    'class',
                                                    'enctype',
                                                    'csrf',
                                                    'id')
            style = style or form_style
            template = template or form_template
            action = action or form_action
            html_id = _id or form_id
            if not self.has_parameter('method'):
                method = form_method
            if action is None:
                action = context['.request.path']

            context['_return'] = form = Form(form_element,
                                             context,
                                             app,
                                             style=style,
                                             template=template,
                                             action=action,
                                             method=method,
                                             enctype=enctype,
                                             _class=_class,
                                             csrf=csrf,
                                             html_id=html_id,
                                             legend=legend)
            context['_content'] = form

            extends = [form_element]
            el = form_element
            visited = {}

            while 1:
                extend = el.extends(context)
                if not extend:
                    break
                parent_el = el
                app, el = el.get_element(extend)
                if (app, el.libid) in visited:
                    raise errors.ElementError("Recursive form extends, element {} has already been extended".format(el),
                                              element=self,
                                              diagnosis="This form has circular references. i.e. the extend attribute references a form that has already been extended.")
                visited[(app, el.libid)] = parent_el
                if not style:
                    style = el.style(context)
                if not template:
                    template = el.template(context)
                if not action:
                    action = el.action(context)
                extends.append(el)
            form.style = style or "simple"
            form.template = template
            form.action = action or ""

            with push_content(context, form.content):
                with form.content.section('fields'):
                    if csrf:
                        form.add_csrf()
                    for el in reversed(extends):
                        yield logic.DeferNodeContents(el)

        if call.has_return:
            form = call.return_value
            bindings = []
            initial_binding = form.get_initial_binding(context)
            if initial_binding:
                bindings.append(initial_binding)
            if src and context[src]:
                bindings.append(form.get_src_binding(context, src))
            if bind and not blank:
                binding = form.get_binding(context, bind)
                bindings.append(binding)
            if bindings:
                if bind is None:
                    form.set_data(context, *bindings)
                else:
                    form.bind(context, *bindings)
        else:
            form = None
        self.set_context(context, dst, form)


class Validate(LogicElement):
    """
    Validates a form

    If the form is passes all of the validate tests in the form, the Attribute
    'ok' will be True and the containing logic will be execute.

    """
    xmlns = namespaces.forms

    src = Attribute("Form source", type="index", default="form", evaldefault=True, map_to="src")
    csrf = Attribute("Enable CSRF check?", type="boolean", default=True)

    class Meta:
        is_call = True
        is_loop = True

    class Help:
        synopsis = "validate a form"

    def logic(self, context):
        form = self.src(context)
        if not form.bound:
            return

        extends = [form.element]
        el = form.element
        while 1:
            extend = el.extends(context)
            if not extend:
                break
            app, el = el.get_element(extend)
            extends.append(el)

        values = {name: field.value for name, field in iteritems(form.fields_map)}
        app = self.get_app(context)
        new_values = {}

        if self.csrf(context):
            form.validate_csrf(context)

        for field in form.fields:
            new_values[field.name] = field.value
            if field.required and field.value in ('', None):
                form.add_fail(field.name, "This field is required")
            else:
                for validate_field in form.get_field_validators(field.name):
                    with self.closure_call(context,
                                           app,
                                           validate_field.data,
                                           _field=field,
                                           form=form,
                                           value=values[field.name],
                                           values=values,
                                           field=field.value):
                        yield logic.DeferNodeContents(validate_field.element)

        if not form.errors:
            for field in form.fields:
                if field.adapt_value is not None:
                    value = new_values[field.name]
                    new_value = field.adapt_value(context, value=value)
                    new_values[field.name] = new_value

            for field in form.fields:
                for adapt_field in form.get_field_adapters(field.name):
                    value = new_values[field.name]
                    with self.closure_call(context,
                                           app,
                                           adapt_field.data,
                                           form=form,
                                           values=values,
                                           value=value) as call:
                        yield logic.DeferNodeContents(adapt_field.element)

                    if call.has_return:
                        new_values[field.name] = call.return_value

        for field in form.fields:
            form.set_field_data(field.name, new_values[field.name])

        form.validated = True
        if form.ok:
            #with self.call(context, app, **context.obj) as call:
            yield logic.DeferNodeContents(self)
            yield logic.SkipNext((namespaces.default, "else"))
            if '_return' in context:
                context['_return'] = call.return_value
                raise logic.Unwind()


class ValidatePost(Validate):
    """
    Validate a form if the current request is a POST request.

    This tag does the same thing as [tag]validate[/tag], but only if the current request is a POST request. Otherwise, the enclosed block will be skipped.

    """

    class Help:
        synopsis = "validate a form for post requests"

    def logic(self, context):
        if context.get('.request.method', '').lower() == 'post':
            for el in super(ValidatePost, self).logic(context):
                yield el


class Apply(LogicElement):
    """
    Applying a form will copy data from a form (in [c]src[/c]) to an object ([c]dst[/c]).

    The name of the values to copy are specified in the [c]dst[/c] attribute of field attributes.

    """
    xmlns = namespaces.forms

    class Meta:
        is_call = True

    class Help:
        synopsis = "apply a form to an object"
        example = """
        <forms:apply src="form" dst="post" />
        """

    src = Attribute("form", type="index", default="form", evaldefault=True)
    dst = Attribute("Destination object", type="reference", required=True, default=None)
    fields = Attribute("fields to apply", type="commalist", default=None)

    def logic(self, context):
        form, dst = self.get_parameters(context, 'src', 'dst')
        if not isinstance(form, Form):
            self.throw('bad-value.form',
                       'form attribute should be a form, not {}'.format(context.to_expr(form)))
        form_data = form.data
        dst_obj = context[dst]
        if not hasattr(dst_obj, 'items'):
            self.throw('moya.forms.bad-dst',
                       "Object referenced by 'dst' must be dict or other mapping type (not {})".format(to_expression(context, dst_obj)))
        if dst:
            field_names = self.fields(context)
            if field_names is None:
                fields = list(form.fields)
            else:
                fields_map = form.fields_map
                fields = [fields_map[name] for name in field_names]

            for field in fields:
                applyers = form.field_applyers[field.name]
                if applyers:
                    for apply_field in applyers:
                        with self.closure_call(context,
                                               form.app,
                                               apply_field.data,
                                               form=form,
                                               object=dst_obj,
                                               values=form.data,
                                               value=form.get_data(field.name)) as call:
                            yield logic.DeferNodeContents(apply_field.element)

                else:
                    with context.frame(dst):
                        field_dst = field.dst
                        if field_dst:
                            value = form_data.get(field.name, None)
                            try:
                                context[field_dst] = value
                            except Exception as e:
                                diagnosis_msg = "The following error was reported: {error}.\n\nCheck you are setting this field to an appropriate value."
                                self.throw('moya.forms.apply-fail',
                                           "unable to set field '{}' to {}".format(field_dst, context.to_expr(value)),
                                           diagnosis=diagnosis_msg.format(error=e),
                                           info={'field': field.name, 'error': text_type(e)})


class FormElement(LogicElement):
    """
    Defines a form.

    """
    xmlns = namespaces.forms

    class Help:
        synopsis = "begin a form definition"
        example = """
        <form libname="form.login" legend="Please login" xmlns="http://moyaproject.com/forms">
            <input name="username" label="Username" type="text" maxlength="30" required="yes"/>
            <password name="password" label="Password" maxlength="30" required="yes"/>
            <submit-button text="Login" />
        </form>
        """

    legend = Attribute("legend text shown above form")
    style = Attribute("Form style", default=None)
    template = Attribute("Form template", default=None)
    action = Attribute("Form action", default=None)
    method = Attribute("Form method", default="post", choices=['get', 'post'])
    enctype = Attribute("Form encoding type", default=None)
    extends = Attribute("Extend another form", default=None)
    _class = Attribute("CSS class override", required=False, default=None)
    csrf = Attribute("Enable csrf protection?", type="boolean", default=True)
    _id = Attribute("Form HTML id", type="text", default=None)

    class Meta:
        tag_name = "form"


@implements_to_string
class FieldElement(LogicElement):
    xmlns = namespaces.forms

    class Help:
        undocumented = True

    enctype = None

    name = Attribute("Field name", required=False, default=None)
    label = Attribute("Label", required=False)
    hidelabel = Attribute("Hide label?", type="boolean", required=False, default=False)
    src = Attribute("Source object index", default=None)
    dst = Attribute("Destination object index", default=None)
    initial = Attribute("Initial value", type="expression", required=False, default=None)
    required = Attribute("Is a value required for this field?", type="boolean")
    help = Attribute("Help text", required=False, default=None)
    inlinehelp = Attribute("Help text", required=False, default=None)
    template = Attribute("Template", type="template", required=False, default=None)
    visible = Attribute("Visible?", type="boolean", required=False, default=True)
    style = Attribute("Override style", required=False, default=None)
    maxlength = Attribute("Maximum length", required=False, default=None)
    adapt = Attribute("Function to adapt field before applying", type="function", default="value", evaldefault=True)
    process = Attribute("Function to process src in to a string", type="function", default="str:value", evaldefault=True)
    disabled = Attribute("Disabled control?", type="boolean", required=False, default=False)
    upload = Attribute("Contains a file upload field?", type="boolean", required=False, default=False)

    def __str__(self):
        return "<%s>" % self._tag_name

    def process_value(self, context, value=None):
        return value

    def get_default(self):
        return None

    def get_field_parameters(self, context):
        params = self.get_all_parameters(context)
        if 'name' in params:
            if params['src'] is None:
                params['src'] = params['name']
            if params['dst'] is None:
                params['dst'] = params['name']
        return params

    def logic(self, context):
        form = context['_return']
        params = self.get_field_parameters(context)
        template = params.pop('template', None)
        style = params.pop('style')

        if params['upload']:
            enctype = "multipart/form-data"
        else:
            enctype = self.enctype

        field = form.add_field(params,
                               enctype=enctype,
                               default=self.get_default(),
                               style=style,
                               process_value=self.process(context),
                               template=template,
                               adapt_value=self.adapt(context),
                               data=self.get_let_map(context))

        content = context['.content']
        context['field'] = field
        if field is not None:
            content.add_renderable(self._tag_name, field)
            with content.node():
                yield logic.DeferNodeContents(self)


class Group(LogicElement):
    """Renders a group of form elements"""
    xmlns = namespaces.forms

    class Help:
        synopsis = "render a group of form elements"

    style = Attribute("Style for child fields", required=False, default=None)

    def logic(self, context):
        form = context['_return']
        params = self.get_parameters(context)
        content = context['.content']

        style = params.style or form.style or 'simple'
        template = "/moya.forms/styles/%s/group.html" % style
        td = {}

        content.add_template("group", template, td)
        with content.node():
            yield logic.DeferNodeContents(self)


class _Field(FieldElement):
    """
    This tag adds a field to the form object, but defers the rendering to enclosed content.

    """

    fieldname = Attribute("Field name", required=False, default="field")

    class Meta:
        synopsis = "add a custom field to a form"
        tag_name = "field"
        example = """
        <field name="example">
            <moya:node template="custom_field.html" />
        </field>
        """

    class Help:
        synopsis = "add a custom field to a form"


class SubmitButton(FieldElement):
    """
    Add a submit button to a form.

    """

    class Help:
        synopsis = "add a submit button in a form"

    fieldname = Attribute("Field name", required=False, default="submit-button")
    name = Attribute("Field name", required=False)
    _class = Attribute("Extra class(es)", required=False, default=None)
    visual = Attribute("Button style", required=False, default="primary")
    block = Attribute("Block level?", required=False, default=False)
    text = Attribute("Text on button", required=False, default="Submit")
    clicked = Attribute("Value when button is clicked", default=None, required=False)


class HiddenInput(FieldElement):
    """
    Add a hidden input to a form.

    """

    class Help:
        synopsis = "add a hidden input value to a form"

    fieldname = Attribute("Field name", required=False, default="hidden")
    visible = Attribute("Visible", type="boolean", required=False, default=False)


class Input(FieldElement):
    """
    Add an input to a form.

    """

    class Help:
        synopsis = "add an input to a form"
        example = """
        <input name="username" maxlength="30" />
        """

    fieldname = Attribute("Field name", required=False, default="text")
    type = Attribute("Input type", required=False, default="text")
    placeholder = Attribute("Placeholder text", required=False, default=None)
    _class = Attribute("Extra class for input", required=False, default="input-medium")


class Upload(FieldElement):
    """
    Add a file upload input to a form.

    """

    class Help:
        synopsis = "add a file upload to a form"

    enctype = "multipart/form-data"

    fieldname = Attribute("Field name", required=False, default="upload")
    type = Attribute("Input type", required=False, default="file")
    placeholder = Attribute("Placeholder text", required=False, default=None)
    _class = Attribute("Extra class for input", required=False, default="")

    def adapt(self, context):

        def process_field_storage(context, value=None):
            if not isinstance(value, FieldStorage):
                return value
            return UploadFileProxy(value)

        return process_field_storage


class Password(FieldElement):
    """
    Add a password input to a form

    """

    class Help:
        synopsis = "add a password field to a form"

    fieldname = Attribute("Field name", required=False, default="password")
    placeholder = Attribute("Placeholder text", required=False, default=None)
    _class = Attribute("Extra class for input", required=False, default="input-medium")


class TextArea(FieldElement):
    """
    Add a textarea (multi-line text input) to a form.

    """

    class Help:
        synopsis = "add a textarea to a form"

    fieldname = Attribute("Field name", required=False, default="text-area")
    rows = Attribute("Number of rows", required=False, default=8, type="integer")
    _class = Attribute("Extra class for input", required=False, default="input-block-level")


class Checkbox(FieldElement):
    """
    Add a checkbox to a form.

    """

    class Help:
        synopsis = "add a checkbox to a form"

    fieldname = Attribute("Field name", required=False, default="check-box")
    on = Attribute("Value when checked", required=False, default="on")
    text = Attribute("Text associated with checkbox", required=False, default='')

    adapt = Attribute("Function to adapt field before applying", type="function", default="value=='on'", evaldefault=True)
    process = Attribute("Function to process src in to a string", type="function", default="value ? 'on' : ''", evaldefault=True)

    def get_default(self):
        return ''


class Radio(FieldElement):
    """
    Add a radio button to a form.

    The user will be able to select just one radio button with the same [c]name[/c] attribute.

    """

    class Help:
        synopsis = "add a radio button to a field"

    fieldname = Attribute("Field name", required=False, default="radio")
    text = Attribute("Text associated with checkbox", required=False, default='', map_to="rtext")
    on = Attribute("Value when selected", required=True)

    def logic(self, context):
        if '_radiogroup' not in context:
            return super(Radio, self).logic(context)
        radiogroup = context['_radiogroup']
        on = self.on(context)
        text = self.rtext(context) or context.sub(self.text)
        radiogroup.add_option(on or text, text)


class RadioGroup(FieldElement):
    """
    Add a radio group (container for radio buttons) to a form.

    """

    fieldname = Attribute("Field name", required=False, default="radio-group")
    choices = Attribute("Possible choices", type="expression", required=False, default=None)
    inline = Attribute("Display inline?", type="boolean", required=False, default=False)

    class Help:
        synopsis = "add a radio group to a form"
        example = """
        <radio-group name="option" label="Radio Group" required="yes">
            <radio text="Option 1" on="1"/>
            <radio text="Option 2" on="2"/>
            <radio text="Option 3" on="3"/>
        </radio-group>
        """

    @classmethod
    def add_choices(cls, select, select_choices):
        for group, choices in select_choices:
            if choices:
                if isinstance(choices, text_type):
                    select.add_option(group, choices)
                else:
                    for choice, choice_label in choices:
                        select.add_option(choice, choice_label, group=group)

    def logic(self, context):
        form = context['_return']
        params = self.get_field_parameters(context)
        template = params.pop('template', None)
        renderable = context['_radiogroup'] = form.add_field(params,
                                                             template=template)

        select_choices = self.choices(context)

        if select_choices:
            self.add_choices(renderable, select_choices)

        yield logic.DeferNodeContents(self)
        del context['_radiogroup']
        context['.content'].add_renderable(self._tag_name, renderable)


class Select(FieldElement):
    """
    Add a select input to a form.

    """

    class Help:
        synopsis = "add a select box to a form"

    fieldname = Attribute("Field name", required=False, default="select")
    _class = Attribute("Extra class for select", required=False, default="input-medium")
    multiple = Attribute("Multiple select?", required=False, default=False)
    choices = Attribute("Possible choices", type="expression", required=False, default=None)

    @classmethod
    def add_choices(cls, select, select_choices):
        for group, choices in select_choices:
            if isinstance(choices, text_type):
                select.add_option(group, choices)
            else:
                for choice, choice_label in choices:
                    select.add_option(choice, choice_label, group=group)

    def logic(self, context):
        form = context['_return']
        params = self.get_field_parameters(context)
        template = params.pop('template', None)
        select = context['_select'] = form.add_field(params, template=template)

        select_choices = self.choices(context)

        if select_choices:
            self.add_choices(select, select_choices)

        yield logic.DeferNodeContents(self)
        del context['_select']
        context['.content'].add_renderable(self._tag_name, select)


class AddChoices(LogicElement):
    """
    Add choices to a [tag]select[/tag] or [tag]check-select[/tag].

    """

    xmlns = namespaces.forms

    class Help:
        synopsis = "add choices to a select tag"

    choices = Attribute("Possible choices", type="expression", required=False, default=None)

    def logic(self, context):
        select = context.get('_select', None)
        if select is None:
            self.throw('add-choices.no-select',
                       'this tag must appear inside a <select>')
        Select.add_choices(select, self.choices(context))


class CheckSelect(FieldElement):
    """
    Add a [i]check select[/i] to a form.

    A check select is an alternative to a multiple select ([tag forms]select[/tag]) which uses checkboxes rather than an input.

    """

    fieldname = Attribute("Field name", required=False, default="check-select")
    _class = Attribute("Extra class for select", required=False, default="input-medium")
    choices = Attribute("Possible choices", type="expression", required=False, default=None)

    class Help:
        synopsis = "add a check select control to a form"

    def logic(self, context):
        form = context['_return']
        params = self.get_field_parameters(context)
        params['multiple'] = True
        template = params.pop('template', None)
        select = context['_select'] = form.add_field(params, template=template)

        select_choices = self.choices(context)
        if select_choices:
            Select.add_choices(select, select_choices)

        yield logic.DeferNodeContents(self)
        del context['_select']
        context['.content'].add_renderable(self._tag_name, select)


class Option(LogicElement):
    """
    Define an option in a select input.

    """
    xmlns = namespaces.forms
    value = Attribute("Value", default=None, required=False)
    group = Attribute("Group", required=False, default=None)
    selected = Attribute("Selected", type="expression", required=False, default=False)
    help = Attribute("Help text", required=False, default=None)
    renderable = Attribute("Renderable object", type="expression", required=False, default=None)

    class Help:
        synopsis = "an option in a select control"
        example = """
        <select name="fruit">
            <option value="apples">Apples</option>
            <option value="oranges">Oranges</option>
            <option value="pears">Pears</option>
        </select>

        """

    def logic(self, context):
        if '_select' not in context:
            # TODO: Throw an error?
            return
        text = context.sub(self.text.strip())
        if self.has_parameter('value'):
            value = self.value(context)
        else:
            value = text
        params = self.get_parameters(context)
        context['_select'].add_option(value,
                                      text,
                                      group=params.group,
                                      help=params.help,
                                      renderable=params.renderable)


class OptGroup(LogicElement):
    """
    Add an optgroup (heading for options) to a select.

    """

    xmlns = namespaces.forms

    label = Attribute("Label", required=True)

    class Help:
        synopsis = "add an optgroup to a select"

    def logic(self, context):
        if '_select' not in context:
            # TODO: Throw an error?
            return

        context['_select'].set_group(self.label(context))
        yield logic.DeferNodeContents(self)
        context['_select'].set_group(None)


class ValidateField(LogicElement):
    """
    Define validation code for a field in a form.

    """
    xmlns = namespaces.forms

    field = Attribute("Field name", required=True, default=None)

    class Meta:
        is_call = True

    class Help:
        synopsis = "validate a field in a form"
        example = """
        <validate-field field="price">
            <fail if="value lte 0>
                Price must be greater than 0.00.
            </fail>
        </validate-field>

        """

    def logic(self, context):
        form = context['_return']
        form.add_field_validator(self.field(context), self.get_closure(context))


class AdaptField(LogicElement):
    xmlns = namespaces.forms

    field = Attribute("Field name", required=True, default=None)

    class Help:
        synopsis = "adapt a field in a form"

    def logic(self, context):
        form = context['_return']
        params = self.get_let_map(context)
        form.add_field_adapter(self.field(context),
                               self.get_closure(context, extra=params))


class ApplyField(LogicElement):
    """
    Apply a form field.

    Invoked by [tag forms]apply[/tag].

    """
    xmlns = namespaces.forms

    class Help:
        synopsis = "apply a form field"

    field = Attribute("Field name", required=True)

    def logic(self, context):
        form = context['_return']
        params = self.get_let_map(context)
        form.add_field_applyer(self.field(context),
                               self.get_closure(context, extra=params))


class Fail(LogicElement):
    """
    Used in a [tag forms]validate-feld[/tag], this tag marks a field as having failed validation, and sets an error message.

    Note that this tag acts like a [tag]return[/tag].

    """

    xmlns = namespaces.forms

    class Meta:
        translate = True

    class Help:
        synopsis = "set a fail in a form validation"
        example = """
        <validate-field field="price">
            <fail if="value lte 0>
                Price must be greater than 0.00.
            </fail>
        </validate-field>
        """

    def logic(self, context):
        field = context['_field'].name
        context['form'].add_fail(field, context.sub(self.text.strip()))
        raise logic.Unwind()


class Actions(LogicElement):
    """
    This tag a group of [tag forms]submit-button[/tag] or other controls that will be rendered as a group (typically a single row).

    """
    xmlns = namespaces.forms

    class Help:
        synopsis = "mark a group of action buttons in a form"
        example = """
        <actions>
            <submit-button name="action" clicked="delete" text="Delete"/>
            <submit-button name="action" clicked="cancel" text="Cancel"/>
        </actions>
        """

    def logic(self, context):
        form = context['_return']
        path = "/moya.forms/styles/%s/actions.html" % form.style
        form.content.add_template(self._tag_name, path, {})

        with form.content.node():
            yield logic.DeferNodeContents(self)


class Error(LogicElement):
    """
    Set an error message on either a field or the entire form.

    """
    xmlns = namespaces.forms

    src = Attribute("Form", default="form", evaldefault=True, type="expression")
    field = Attribute("Field name", required=False, default=None)

    class Meta:
        translate = True

    class Help:
        synopsis = "add an error message to a form"
        example = """
        <forms:error src="form" if=".user">
            You must be logged in to do that!
        </forms:error>

        """

    def logic(self, context):
        field = self.field(context)
        form = self.src(context)
        text = context.sub(self.text)
        if field is None:
            form.error = text
        else:
            form.errors[field] = text



from __future__ import unicode_literals, print_function

from ..errors import BadValueError
from ..versioning import VersionSpec
from ..context import Expression, TrueExpression, FalseExpression, dataindex
from ..dbexpression import DBExpression
from ..context.expressiontime import TimeSpan
from ..context.color import Color as ExpressionColor
from ..context.tools import to_expression
from ..elements.elementproxy import ElementProxy
from ..http import StatusCode
from ..compat import (implements_to_string,
                      int_types,
                      text_type,
                      string_types,
                      with_metaclass)


__all__ = ["Constant",
           "Text",
           "Number",
           "Integer",
           "Index",
           "ExpressionAttribute",
           "ApplicationAttribute",
           "Reference",
           "TimeSpanAttribute",
           "CommaList"]


valid_types = []


class AttributeTypeMeta(type):
    """Keeps a registry of all AttributeType classes"""
    registry = {}

    def __new__(cls, name, base, attrs):
        new_class = type.__new__(cls, name, base, attrs)
        if name != "AttributeType":
            name = getattr(new_class, 'name', name.lower())
            AttributeTypeMeta.registry[name] = new_class
            valid_types.append(name)
        return new_class


def lookup(name):
    """Return the AttributeType of a given name

    Non-string are passed through, unaltered.

    """
    if not isinstance(name, string_types):
        return name
    try:
        return AttributeTypeMeta.registry[name.lower()]
    except KeyError:
        raise KeyError(name)


@implements_to_string
class AttributeTypeBase(object):
    """Base class for attribute types"""
    translate = False

    def __init__(self, element, attribute_name, text):
        self.attribute_name = attribute_name
        self.element = element
        self.text = text
        if not isinstance(text, string_types):
            self.const = True
        else:
            self.const = '${' not in text
        super(AttributeTypeBase, self).__init__()

    def __call__(self, context):
        if self.const:
            return self.process(self.text)
        return self.process(context.sub(self.text))

    @classmethod
    def check(self, value):
        return None

    @property
    def value(self):
        if not hasattr(self, '_value'):
            self._value = self.process(self.text)
        return self._value

    @classmethod
    def get_type_display(cls):
        return cls.type_display

    @classmethod
    def display(cls, value):
        if isinstance(value, text_type):
            return '"{}"'.format(value)

    def process(self, text):
        return text

    def __str__(self):
        return self.text

    def invalid(self, text):
        raise BadValueError("In attribute '{}': '{}' is not a valid {}".format(self.attribute_name, text, self.get_type_display()))


class AttributeType(with_metaclass(AttributeTypeMeta, AttributeTypeBase)):
    pass


class Constant(AttributeType):
    type_display = "constant"

    def __init__(self, element, attribute_name, value):
        self.element = element
        self.attribute_name = attribute_name
        self.value = value

    def __call__(self, context):
        return self.value

    @classmethod
    def get_type_display(self):
        return text_type(self.value)


class Text(AttributeType):
    type_display = "text"

    def __call__(self, context):
        if self.const:
            return self.text
        return self.process(context.sub(self.text))


class Bytes(AttributeType):
    type_display = "bytes"

    def process(self, text):
        try:
            text = self.text.encode('utf-8')
        except:
            raise BadValueError("must be encodeable as UTF-8")
        return text


class Raw(AttributeType):
    type_display = "raw"

    def __call__(self, context):
        return self.text


class Number(AttributeType):
    type_display = "number"

    def process(self, text):
        try:
            return float(text)
        except ValueError:
            self.invalid(text)

    @classmethod
    def check(cls, value):
        if '${' not in value:
            try:
                float(value)
            except:
                return 'expected a number, not "{}"'.format(value)


class Integer(AttributeType):
    type_display = "integer"

    def process(self, text):
        try:
            return int(float(text))
        except ValueError:
            self.invalid(text)

    @classmethod
    def display(cls, value):
        if isinstance(value, int_types):
            return text_type(value)

    @classmethod
    def check(cls, value):
        if '${' not in value:
            try:
                int(float(value))
            except:
                return 'expected an integer, not "{}"'.format(value)


class Color(AttributeType):
    type_display = "color"

    def __call__(self, context):
        return ExpressionColor.parse(context.sub(self.text))

    @classmethod
    def check(cls, value):
        if '${' not in value:
            try:
                ExpressionColor.parse(value)
            except Exception as e:
                return text_type(e)


class Index(AttributeType):
    type_display = "index"

    def __call__(self, context):
        return context.get_sub(self.text)


class Reference(AttributeType):
    type_display = "reference"

    def __init__(self, element, attribute_name, text):
        super(Reference, self).__init__(element, attribute_name, text)
        self.reference = dataindex.parse(self.text)

    def __call__(self, context):
        return self.reference


class Element(AttributeType):
    type_display = "element"

    def __call__(self, context):
        element_ref = context.sub(self.text)
        app, element = self.element.archive.get_element(element_ref, app=context.get(".app"))
        return ElementProxy(context, app, element)


class ElementRef(AttributeType):
    type_display = "element reference"

    def __call__(self, context):
        return context.sub(self.text)


class ExpressionAttribute(AttributeType):
    type_display = "expression"
    name = "expression"

    def __init__(self, element, attribute_name, text):
        self.element = element
        self.attribute_name = attribute_name
        self.exp = Expression(text)

    def __call__(self, context):
        return self.exp.eval(context)

    @classmethod
    def display(cls, value):
        return to_expression(None, value)

    @classmethod
    def check(self, value):
        try:
            Expression(value).compile()
        except Exception as e:
            return text_type(e)
        else:
            return None


class FunctionAttribute(AttributeType):
    type_display = "function"
    name = "function"

    def __init__(self, element, attribute_name, text):
        self.element = element
        self.attribute_name = attribute_name
        self.exp = Expression(text)

    def __call__(self, context):
        return self.exp.make_function(context)

    @classmethod
    def check(self, value):
        try:
            Expression(value).compile()
        except Exception as e:
            return text_type(e)
        else:
            return None


class ApplicationAttribute(AttributeType):
    type_display = "application reference"
    name = "application"

    def __init__(self, element, attribute_name, text):
        self.element = element
        self.attribute_name = attribute_name
        self.text = text

    def __call__(self, context):
        app_name = context.sub(self.text)
        if app_name == '':
            raise BadValueError("app name must not be an empty string")
        if not app_name:
            return None
        try:
            app = self.element.archive.find_app(app_name)
        except Exception as e:
            raise BadValueError(text_type(e))
        return app


class DBExpressionAttribute(AttributeType):
    type_display = "database expression"
    name = "dbexpression"

    def __init__(self, element, attribute_name, text):
        self.element = element
        self.attribute_name = attribute_name
        self.text = text

    def __call__(self, context):
        text = context.sub(self.text)
        return DBExpression(text) if text else None

    @classmethod
    def check(self, value):
        if '${' not in value:
            try:
                DBExpression(value).compile()
            except Exception as e:
                return text_type(e)
            else:
                return None


class Boolean(ExpressionAttribute):
    type_display = "boolean"
    name = "boolean"

    def __init__(self, element, attribute_name, text):
        self.attribute_name = attribute_name
        self.text = text
        if text in ('yes', 'True'):
            self.exp = TrueExpression()
        elif text in ('no', 'False'):
            self.exp = FalseExpression()
        else:
            self.exp = Expression(text)

    @classmethod
    def display(cls, value):
        if isinstance(value, bool):
            return "yes" if value else "no"

    def __call__(self, context):
        return bool(self.exp.eval(context))

    @classmethod
    def check(self, value):
        try:
            Expression(value).compile()
        except Exception as e:
            return text_type(e)
        else:
            return None


class DictExpression(ExpressionAttribute):
    type_display = "dict expression"
    name = "dict"

    def __init__(self, element, attribute_name, text):
        self.attribute_name = attribute_name
        self.exp = Expression(text)

    def __call__(self, context):
        d = self.exp.eval(context)
        if not (hasattr(d, 'items') and hasattr(d, '__getitem__')):
            raise BadValueError("must be a dictionary or similar type")
        return d

    @classmethod
    def check(self, value):
        try:
            Expression(value).compile()
        except Exception as e:
            return text_type(e)
        else:
            return None


class TimeSpanAttribute(AttributeType):
    type_display = "timespan"
    name = "timespan"

    def process(self, text):
        return TimeSpan(text)

    @classmethod
    def check(self, value):
        if '${' not in value:
            try:
                TimeSpan(value)
            except Exception as e:
                return text_type(e)
            else:
                return None


class CommaList(AttributeType):
    type_display = "comma list"
    name = "commalist"

    def process(self, text):
        if ',' in text:
            return [t.strip() for t in text.split(',')]
        else:
            return [text]


class Namespace(AttributeType):
    type_display = "namespace"
    name = "namespace"


class Templates(AttributeType):
    type_display = "list of template paths"
    name = "templates"

    def __call__(self, context):
        sub = context.sub
        text = self.text
        if ',' in text:
            return [sub(t.strip()) for t in text.split(',')]
        else:
            return [text.strip()]


class Template(AttributeType):
    type_display = "template path"
    name = "template"

    def __call__(self, context):
        text = context.sub(self.text).strip()
        return text


class Version(AttributeType):
    type_display = "version spec"
    name = "version"

    def __call__(self, context):
        text = VersionSpec(self.text)
        return text

    @classmethod
    def check(self, value):
        try:
            VersionSpec(value)
        except Exception as e:
            return text_type(e)
        else:
            return None


class HTTPStatus(AttributeType):
    type_display = "http status code"
    name = "httpstatus"

    def __call__(self, context):
        if self.text.isdigit():
            return int(self.text)
        try:
            status_code = StatusCode(self.text)
        except KeyError:
            raise BadValueError("'{}' is not a valid status code".format(self.text))
        status = int(status_code)
        return status

    @classmethod
    def check(self, value):
        try:
            StatusCode(value)
        except Exception as e:
            return ValueError("'{}' is not a valid status code".format(value))
        else:
            return None

if __name__ == "__main__":

    from moya.context import Context
    c = Context()
    c['foo'] = [1, 2, 3, 4, 5]
    c['fruit'] = ['apples', 'orange', 'pears']

    t = Number("${foo.2}.14")
    print(t.__call__)
    print(repr(t))
    print(t(c))

    i = Index("fruit.${foo.1}")
    print(i(c))



from __future__ import unicode_literals
from __future__ import print_function

from .compat import quote_plus, text_type


def _iter_qs_map(qs_map):
    for k, v in qs_map.items():
        if isinstance(v, list):
            for _v in v:
                yield k, _v
        else:
            yield k, v


def urlencode(query, _quote_plus=quote_plus, _iter_qs=_iter_qs_map):
    """url encode a mapping of query string values"""
    # Works slightly differently to Pythons urlencode
    # This function accepts lists and values, to generate multiple keys. e.g {'foo': ['bar', 'baz']} -> "foo=bar&foo=baz"
    # It will also remove the = if the value is empty, e.g. {'foo': ''} -> "foo"
    qs = "&".join("{}={}".format(_quote_plus(k), _quote_plus(v)) if text_type(v) else _quote_plus(k)
                  for k, v in _iter_qs(query))
    return qs



from __future__ import unicode_literals
from __future__ import print_function

import argparse
import os
import sys
import io
import glob
import tempfile

from moya.compat import text_type
from moya.settings import SettingsContainer
from moya.console import Console

DEFAULT_HOME_DIR = '/etc/moya/'


class CommandError(Exception):
    pass


class MoyaSrv(object):
    """Moya Service"""

    def get_argparse(self):
        parser = argparse.ArgumentParser(prog="moya-srv",
                                         description=self.__doc__)

        parser.add_argument('-d', '--debug', dest="debug", action="store_true",
                            help="enable debug information (show tracebacks)")
        parser.add_argument('--home', dest="home", metavar="PATH", default=None,
                            help="moya service directory")

        subparsers = parser.add_subparsers(title="available sub-commands",
                                           dest="subcommand",
                                           help="sub-command help")

        list_parser = subparsers.add_parser('list',
                                            help="list projects",
                                            description="list enabled projects")
        list_parser

        where_parser = subparsers.add_parser('where',
                                             help='find project directory',
                                             description="output the location of the project")

        where_parser.add_argument(dest="name", metavar="PROJECT",
                                  help="name of a project")

        restart_parser = subparsers.add_parser('restart',
                                               help='restart a project server',
                                               description="restart a server")

        restart_parser.add_argument(dest="name", metavar="PROJECT",
                                    help="name of a project")

        return parser

    def error(self, msg, code=-1):
        sys.stderr.write(msg + '\n')
        sys.exit(code)

    def run(self):
        parser = self.get_argparse()
        self.args = args = parser.parse_args(sys.argv[1:])
        self.console = Console()

        self.home_dir = args.home or os.environ.get('MOYA_SRV_HOME', None) or DEFAULT_HOME_DIR

        settings_path = os.path.join(self.home_dir, 'moya.conf')
        try:
            with io.open(settings_path, 'rt') as f:
                self.settings = SettingsContainer.read_from_file(f)
        except IOError:
            self.error('unable to read {}'.format(settings_path))
            return -1

        method_name = "run_" + args.subcommand.replace('-', '_')
        try:
            return getattr(self, method_name)() or 0
        except CommandError as e:
            self.error(text_type(e))
        except Exception as e:
            if args.debug:
                raise
            self.error(text_type(e))

    def _get_projects(self):
        project_paths = self.settings.get_list('projects', 'read')
        paths = []
        cwd = os.getcwd()
        try:
            os.chdir(self.home_dir)
            for path in project_paths:
                glob_paths = glob.glob(path)
                paths.extend([os.path.abspath(p) for p in glob_paths])
        finally:
            os.chdir(cwd)
        return paths

    def project_exists(self, name):
        for path in self._get_projects():
            settings = self.read_project(path)
            if settings.get('service', 'name', None) == name:
                return True
        return False

    def read_project(self, path):
        settings = SettingsContainer.read_os(path)
        return settings

    def run_list(self):
        table = []
        for path in self._get_projects():
            settings = self.read_project(path)
            location = settings.get('service', 'location', '?')
            name = settings.get('service', 'name', '?')
            domains = "\n".join(settings.get_list('service', 'domains', ""))
            table.append([name, domains, path, location])
        self.console.table(table, header_row=['name', 'domain(s)', 'conf', 'location'])

    def run_where(self):
        name = self.args.name
        location = None
        for path in self._get_projects():
            settings = self.read_project(path)
            if settings.get('service', 'name', None) == name:
                location = settings.get('service', 'location', None)
        if location is None:
            self.error("no project '{}'".format(name))
        sys.stdout.write(location)

    def run_restart(self):
        name = self.args.name
        if not self.project_exists(name):
            self.error("no project '{}'".format(name))
        temp_dir = os.path.join(self.settings.get('service', 'temp_dir', tempfile.gettempdir()), 'moyasrv')
        try:
            os.makedirs(temp_dir)
        except OSError:
            pass
        change_path = os.path.join(temp_dir, "{}.changes".format(name))
        try:
            with open(change_path, 'a'):
                os.utime(change_path, None)
        except IOError as e:
            sys.stderr.write("{}\n".format(text_type(e)))
            return -1


def main():
    moya_srv = MoyaSrv()
    sys.exit(moya_srv.run() or 0)



from __future__ import unicode_literals, print_function

from .context.missing import is_missing
from .compat import text_type


class MoyaFilterBase(object):

    def __init__(self, value_name="value", allow_missing=False):
        self._value_name = value_name
        self.allow_missing = allow_missing
        super(MoyaFilterBase, self).__init__()

    def get_value_name(self):
        return self._value_name

    def __moyabind__(self, app):
        return BoundFilter(app, self)

    def __moyafilter__(self, context, app, value, params):
        if not self.allow_missing and is_missing(value):
            raise ValueError("{} doesn't accept a missing value (left hand side is {!r})".format(self, value))

        params[self.get_value_name()] = value
        filter_call = self.lib.archive.get_callable(self.element_ref, app=app)
        value = filter_call(context, **params)
        return value

    def __moyacall__(self, params):
        return MoyaFilterParams(self, params)


class BoundFilter(MoyaFilterBase):
    def __init__(self, _app, _filter):
        self._app = _app
        self._filter = _filter
        self.validator = _filter.validator
        super(BoundFilter, self).__init__()

    def get_value_name(self):
        return self._filter.get_value_name()

    def __repr__(self):
        return "{!r} (app is '{}')".format(self._filter, self._app.name)

    def __moyafilter__(self, context, app, value, params):
        params['_caller_app'] = app
        if self.validator is not None:
            self.validator.check(context, params, self)
        return self._filter.__moyafilter__(context, self._app, value, params)


class MoyaFilter(MoyaFilterBase):
    def __init__(self, lib, filter_element, value_name, allow_missing=False, validator=None):
        self.lib = lib
        self.element_ref = filter_element
        self.validator = validator
        super(MoyaFilter, self).__init__(value_name, allow_missing=allow_missing)

    def __repr__(self):
        return "<filter '{}'>".format(self.element_ref)


class MoyaFilterExpression(MoyaFilterBase):
    def __init__(self, exp, value_name, allow_missing=False):
        self.exp = exp
        super(MoyaFilterExpression, self).__init__(value_name, allow_missing=allow_missing)

    def __repr__(self):
        return "<filter '{}'>".format(text_type(self.exp))

    def __moyafilter__(self, context, app, value, params):
        if is_missing(value):
            return value
        params[self.get_value_name()] = value
        return self.exp(context, params)

    def __moyacall__(self, params):
        return MoyaFilterParams(self, params)


class MoyaFilterParams(object):
    def __init__(self, filter, params):
        self.filter = filter
        self.params = params

    def __repr__(self):
        return repr(self.filter)

    def __moyafilter__(self, context, app, value, params):
        #if self.filter.validator is not None:
        #    self.filter.validator.check(context, params)
        return self.filter.__moyafilter__(context, app, value, self.params)



from __future__ import unicode_literals

from ..elements.elementbase import Attribute
from ..tags.context import ContextElementBase
from .. import namespaces
from .. import logic


class Check(ContextElementBase):
    """A pre-flight check"""
    xmlns = namespaces.preflight

    class Help:
        synopsis = "define a pre-flight test"


class Result(ContextElementBase):
    xmlns = namespaces.preflight
    exit = Attribute("Also exit the check", type="boolean", default=False)
    status = None

    class Help:
        undocumented = True

    def logic(self, context):
        check = self.get_ancestor((self.xmlns, "check"))
        text = context.sub(self.text)
        context['.preflight'].append((check, self.status, text))
        if self.exit(context):
            raise logic.Unwind()


class Pass(Result):
    status = "pass"

    class Help:
        synopsis = "pass a preflight check"


class Fail(Result):
    status = "fail"

    class Help:
        synopsis = "fail a preflight check"


class Warning(Result):
    status = "warning"

    class Help:
        synopsis = "add a warning result to a preflight check"



from __future__ import unicode_literals
from __future__ import print_function

from .compat import PY2, text_type, implements_to_string
from .urltools import urlencode

from threading import Lock

if PY2:
    from urlparse import parse_qsl
else:
    from urllib.parse import parse_qsl


from collections import OrderedDict


class LRUCache(OrderedDict):
    """A dictionary-like container that stores a given maximum items.

    If an additional item is added when the LRUCache is full, the least recently used key is
    discarded to make room for the new item.

    """
    def __init__(self, cache_size=None):
        self.cache_size = cache_size
        self.lock = Lock()
        super(LRUCache, self).__init__()

    def __reduce__(self):
        return self.__class__, (self.cache_size,)

    def __setitem__(self, key, value):
        with self.lock:
            if self.cache_size is not None and key not in self:
                if len(self) >= self.cache_size:
                    self.popitem(last=False)
            OrderedDict.__setitem__(self, key, value)

    def lookup(self, key):
        with self.lock:
            value = OrderedDict.__getitem__(self, key)
            del self[key]
            OrderedDict.__setitem__(self, key, value)
            return value


@implements_to_string
class QueryData(OrderedDict):
    """A container for data encoded in a url query string"""

    @classmethod
    def from_qs(cls, qs, change_callback=None):
        qd = cls()
        for k, v in parse_qsl(qs, keep_blank_values=True, strict_parsing=False):
            qd.setdefault(k, []).append(v)
        return qd

    def copy(self):
        return OrderedDict(self)

    def update(self, d):
        """Specialized update, setting a value to None will delete it. Also ensures that the query data contains lists"""
        for k, v in d.items():
            if v is None:
                if k in self:
                    del self[k]
            else:
                if isinstance(v, (list, set, tuple, dict)) or hasattr(v, 'items'):
                    self[k] = list(v)
                else:
                    if v is None:
                        v = ''
                    elif not isinstance(v, text_type):
                        v = text_type(v)
                    self[k] = [v]

    def __str__(self):
        return urlencode(self)

    def __repr__(self):
        return '<querydata "{}">'.format(urlencode(self))

    def __setitem__(self, k, v):
        if v is None:
            ret = self.__delitem__(k)
        else:
            if isinstance(v, (set, tuple)):
                v = list(v)
            if not isinstance(v, list):
                v = [text_type(v)]
            ret = super(QueryData, self).__setitem__(k, v)
        return ret

    def __delitem__(self, k):
        ret = super(QueryData, self).__delitem__(k)
        return ret


if __name__ == "__main__":
    qd = QueryData.from_qs('foo=bar&a=1&b=2&hobbit=frodo&hobbit=sam')

    print(qd.items())

    qd.update({'foo': None})

    print(qd.items())



from __future__ import unicode_literals
from __future__ import print_function

import unittest

from moya.build import build_server
from moya import db

import os
from os.path import dirname, join, abspath

curdir = dirname(abspath(__file__))

from moya.console import Console
from moya import pilot


class TestDB(unittest.TestCase):

    def setUp(self):

        try:
            os.remove("dbtest.sqlite")
        except OSError:
            pass

        build_result = build_server(join(curdir, 'dbtest/'), "settings.ini")
        archive = build_result.archive
        base_context = build_result.context

        archive.populate_context(build_result.context)
        console = Console()
        db.sync_all(build_result.archive, console)

        context = base_context.clone()
        root = context.root
        root['_dbsessions'] = db.get_session_map(archive)
        archive.call("dbtest#setup", context, None, id=1)
        db.commit_sessions(context)

        #build_result = build_server(join(curdir, 'dbtest/'), "settings.ini")
        self.archive = build_result.archive
        self.base_context = build_result.context

        self.context = self.base_context.clone()
        archive.populate_context(self.context)

        root = self.context.root
        root['_dbsessions'] = db.get_session_map(self.archive)

    def tearDown(self):
        try:
            os.remove("dbtest.sqlite")
        except OSError as e:
            pass

    def test_bulk_create(self):
        """Test bulk create of db"""
        context = self.context
        print(context.root.keys())
        obj = self.archive.call("dbtest#get_by_id", context, None, id=1)
        self.assertEqual(obj.id, 1)

        obj = self.archive.call("dbtest#get_by_id", context, None, id=2)

        self.assertEqual(obj.id, 2)
        obj = self.archive.call("dbtest#get_by_title", context, None, title="Zen 2")
        self.assertEqual(obj.id, 2)

    def test_owner(self):
        """Test object ownership"""
        context = self.context
        call = self.archive.call
        with pilot.manage(context):
            call("dbtest#owner_test", context, None)
            owner = call("dbtest#get_owner", context, None, name="owner")
            print(owner)
            self.assertEqual(owner.name, 'owner')

            child1 = call("dbtest#get_child", context, None, name="child1")
            child2 = call("dbtest#get_child", context, None, name="child2")
            child3 = call("dbtest#get_child", context, None, name="child3")
            child4 = call("dbtest#get_child", context, None, name="child4")

            print(child1, child2, child3, child4)
            assert child1, "child1 should exist"
            assert child2, "child2 should exist"
            assert child3, "child3 should exist"
            assert child4, "child4 should exist"
            assert owner.unowned_child is child1
            assert owner.owned_child is child2
            assert owner.unowned_child_o2o is child3
            assert owner.owned_child_o2o is child4

            call("dbtest#delete_owner", context, None, name="owner")
            print(owner)
            owner = call("dbtest#get_owner", context, None, name="owner")

            print(owner)
            assert not owner, "owner is {}".format(owner)

            child1 = call("dbtest#get_child", context, None, name="child1")
            child2 = call("dbtest#get_child", context, None, name="child2")
            child3 = call("dbtest#get_child", context, None, name="child1")
            child4 = call("dbtest#get_child", context, None, name="child2")

            print(child1, child2, child3, child4)
            assert child1, "child1 should exists"
            assert not child2, "child2 should not exists"
            assert child3, "child3 should exists"
            assert not child4, "child4 should not exists"

    def test_owned(self):
        """Test owned objects"""

        context = self.context
        call = self.archive.call
        with pilot.manage(context):
            call('dbtest#make_post', context, None)

            post = call('dbtest#get_post', context, None, name="post")
            assert post.name == 'post', "post should be called 'post'"
            assert post.images[0].name == 'images1', 'expected images1'

            images = call('dbtest#get_images', context, None, name="images1")
            assert images.name == 'images1', "images should be named 'images1'"

            call('dbtest#delete_post', context, None, name="post")

            images = call('dbtest#get_images', context, None, name="images1")
            assert not images, "images should have been deleted"



from .. import iniparse
from fs.path import dirname, pathjoin


class Page(object):
    def __init__(self, doc_class, settings):
        self.doc_class = doc_class
        self.settings = settings

    def __repr__(self):
        return "Page(%r, %r)" % (self.doc_class, self.settings)

    def get(self, context, settings_name):
        return context.sub(self.settings.get(settings_name, ''))

    def get_path(self, context):
        return context.sub(self.settings.get('path', ''))


class Theme(object):
    def __init__(self, fs):
        self.fs = fs
        self.cfg = None
        self.theme_settings = None
        self.pages = []
        self.read()

    def get(self, section_name, key, default=None):
        section = self.cfg.get(section_name, None)
        if section is None:
            return default
        return section.get(key, default)

    def read(self):
        with self.fs.open('theme.ini', 'rb') as settings_file:
            cfg = iniparse.parse(settings_file)
        self.cfg = cfg

        self.theme_settings = cfg.get('theme', {})

        for section, settings in cfg.items():
            what, _, name = section.partition(':')
            if what == 'page':
                page = Page(name, settings)
                self.pages.append(page)

    def get_pages(self, doc):
        doc_class = doc.doc_class
        for page in self.pages:
            if page.doc_class == doc_class:
                yield page

    def get_relative_path(self, path):
        ini_path = dirname(self.fs.getsyspath('theme.ini'))
        path = pathjoin(ini_path, path)
        return path



"""

An experimental work in progress asynchronous logic executor


"""

import threading


class ExecutorThread(threading.thread):
    def __init__(self):
        pass


class AsyncLogic(object):
    pass



from __future__ import unicode_literals

from ..elements.registry import ElementRegistry
from ..elements.elementbase import LogicElement, DynamicElementMixin
from ..template.moyatemplates import Template
from ..html import escape
from ..render import HTML
from .. import logic
from .. import namespaces


def make_html_tag(_tag_name):

    class DynamicHTMLTag(LogicElement, DynamicElementMixin):
        xmlns = namespaces.html
        template = Template("<${tag_name}${tag_attribs}>{% children %}</${tag_name}>", __file__)

        class Meta:
            text_nodes = "text"
            dynamic = True
            tag_name = _tag_name
            translate = False
            all_attributes = True

        def logic(self, context):
            sub = context.sub
            tag_attribs = " " + " ".join('{}="{}"'.format(escape(k), escape(sub(v)))
                                         for k, v in self._attrs.items()
                                         if k != 'if')
            td = {"tag_attribs": HTML(tag_attribs.rstrip()),
                  "tag_name": HTML(_tag_name)}
            content = context['.content']
            content.add_template(self._tag_name, self.template, td)
            with content.node():
                yield logic.DeferNodeContents(self)
    return DynamicHTMLTag


ElementRegistry.default_registry.add_dynamic_registry(namespaces.html, make_html_tag)



from __future__ import unicode_literals
from __future__ import print_function

from .response import MoyaResponse
from .compat import PY2, py2bytes
from . import http
from .tools import md5_hexdigest
from . import logic

from fs.path import basename
from fs.errors import FSError

from datetime import datetime
import mimetypes
from pytz import UTC


from . import __version__
SERVER_NAME = "Moya/{}.{}".format(*__version__.split('.')[:2])


def serve_file(req, fs, path, filename=None):
    """Serve a static file"""
    res = MoyaResponse()
    mime_type, encoding = mimetypes.guess_type(basename(path))
    if mime_type is None:
        mime_type = b"application/octet-stream" if PY2 else "application/octet-stream"

    # File does not exist
    if not path or not fs.isfile(path):
        raise logic.EndLogic(http.RespondNotFound())

    serve_file = None
    # Get file info
    try:
        file_size = fs.getsize(path)
        info = fs.getinfokeys(path, 'modified_time')
        serve_file = fs.open(path, 'rb')
    except FSError:
        # Files system read failed for some reason
        if serve_file is not None:
            serve_file.close()
        raise logic.EndLogic(http.RespondNotFound())
    else:
        # Make a response
        mtime = info.get('modified_time', None)
        if mtime is None:
            mtime = datetime.utcnow()
        else:
            mtime = UTC.localize(mtime)
        res.date = datetime.utcnow()
        res.content_type = py2bytes(mime_type)
        res.last_modified = mtime
        res.etag = "{}-{}-{}".format(mtime, file_size, md5_hexdigest(path))
        res.server = SERVER_NAME
        if filename is not None:
            res.content_disposition = 'attachment; filename="{}"'.format(filename)

        # Detect 304 not-modified
        status304 = False
        if req.if_none_match and res.etag:
            status304 = res.etag in req.if_none_match
        elif req.if_modified_since and res.last_modified:
            status304 = res.last_modified <= req.if_modified_since
        if status304:
            res.status = 304
            serve_file.close()
        else:
            # Use high performance file wrapper if available
            if 'wsgi.file_wrapper' in req.environ:
                res.app_iter = req.environ['wsgi.file_wrapper'](serve_file)
            else:
                res.body_file = serve_file
        # Set content length
        if not status304:
            res.content_length = file_size
    raise logic.EndLogic(res)



"""Check a Python version specifier against a version of Python

Examples of the version specifier syntax are as follows:

"py"        Any version of Python
"py2"       Any version in the 2.X series
"py2.6"     Only Python version 2.6
"py2.5+"    Python version 2.5 onwards, but only in the Py2.X series
"py3.2+"    Python version 3.2 onwards, but only in the Py3.X series

Version specifiers may be combined with commas. The check function will return True if *any* version passes.

For example:

"py2.5,py2.6"   Matches Python 2.5 and Python 2.6
"py2.7,py3.2+"  Matches Python 2.7 and Python 3.2 onwards

"""
from __future__ import unicode_literals
from __future__ import print_function

import sys
import re


class VersionFormatError(ValueError):
    pass


# Known versions of Python (should be kept up to date)
# Starting from 2.6, since that is the earliest version that Moya requires
known_versions = [(2, 6),
                  (2, 7),
                  (3, 1),
                  (3, 2),
                  (3, 3),
                  (3, 4)]


def check(version_spec, py_version=None, _re_version=re.compile(r'^(py)$|^py(\d)$|^py(\d)\.(\d)$|^py(\d)\.(\d)\+$')):
    """Check a Python version specifier

    `py_version` should be a tuple of (<major version>, <minor version>),
    if it not specified the version of Python will be read from sys.version_info.

    """
    if py_version is None:
        version_info = sys.version_info
        major = version_info.major
        minor = version_info.minor
    else:
        major, minor = py_version
    tokens = [token.strip() for token in version_spec.split(',') if token]
    for version in tokens:
        match = _re_version.match(version)
        if not match:
            raise VersionFormatError("{} is not a valid Py version spec".format(version))
        (all_py,
         absolute_major_version,
         major_version, minor_version,
         major_version_plus, minor_version_plus) = match.groups()
        if (all_py or
           (absolute_major_version and int(absolute_major_version) == major) or
           (major_version and int(major_version) == major and int(minor_version) == minor) or
           (major_version_plus and int(major_version_plus) == major and minor >= int(minor_version_plus))):
                return True
    return False


def list_compatible(version_spec, versions=None):
    """Returns a list of versions compatible with the version spec"""
    if versions is None:
        versions = known_versions
    return [version for version in versions
            if check(version_spec, version)]


if __name__ == "__main__":

    print("Python 2?", check("py2"))
    print("Python 3?", check("py3"))

    tests = ["py", "py2", "py2.6", "py2.6+", "py3"]

    for test in tests:
        print(test, check(test, (2, 7)))

    print(check("py2.5+,py2.6", (2, 7)))
    print(check("py2.7,py3.2+", (3, 3)))

    print(list_compatible('py'))
    print(list_compatible('py2'))
    print(list_compatible('py2.7+,py3'))
    print(list_compatible('py3.2+'))
    print(list_compatible('py2.7,py3.2+'))



from __future__ import unicode_literals

import gettext

import logging
log = logging.getLogger('moya.startup')


class Translations(object):
    def __init__(self):
        self._translations = {}
        self._null = gettext.NullTranslations()

    def read(self, domain, localedir, languages):
        for lang in languages:
            try:
                translations = gettext.translation(domain,
                                                   localedir,
                                                   [lang])
            except IOError:
                log.warning("no translations found for language code '{}'".format(lang))
                translations = None

            if translations is not None:
                self._translations[lang] = translations

    def get(self, languages):
        for lang in languages:
            if lang in self._translations:
                return self._translations[lang]
        return self._null



from ...command import SubCommand
from ...wsgi import WSGIApplication
from ...loggingconf import init_logging
from ...compat import PY2, socketserver

from fs.opener import fsopendir
import os.path
from os.path import join as pathjoin

import sys
from wsgiref.simple_server import (WSGIServer,
                                   WSGIRequestHandler,
                                   make_server)


if PY2:
    from thread import interrupt_main
else:
    from _thread import interrupt_main


import logging
log = logging.getLogger('moya.runtime')


class ThreadedWSGIServer(socketserver.ThreadingMixIn, WSGIServer):
    daemon_threads = True


class RequestHandler(WSGIRequestHandler):

    # Disable simple_server's logging to stdout
    def log_message(self, format, *args):
        pass


class Serve(SubCommand):
    """Serve static files with moya.static"""
    help = "serve static files"

    def add_arguments(self, parser):
        parser.add_argument(dest="fs", metavar="PATH",
                            help="Path to serve")
        parser.add_argument('--host', dest='host', default='127.0.0.1',
                            help="server host")
        parser.add_argument('-p', '--port', default='8000',
                            help="server port")
        parser.add_argument('-t', '--templates', dest="serve_templates", action="store_true",
                            help="render and serve .html files as moya templates")

    def run(self):
        args = self.args
        fs = fsopendir(args.fs)

        from ...command.sub import project_serve
        location = os.path.dirname(project_serve.__file__)

        init_logging(pathjoin(location, 'logging.ini'))

        if args.serve_templates:
            ini = 'templatesettings.ini'
        else:
            ini = 'settings.ini'

        application = WSGIApplication(location, ini, 'main', disable_autoreload=True)
        application.archive.filesystems['static'] = fs

        server = make_server(args.host,
                             int(args.port),
                             application,
                             server_class=ThreadedWSGIServer,
                             handler_class=RequestHandler)
        log.info("server started on http://{}:{}".format(args.host, args.port))

        def handle_error(request, client_address):
            _type, value, tb = sys.exc_info()
            if isinstance(value, KeyboardInterrupt):
                interrupt_main()
        server.handle_error = handle_error

        try:
            server.serve_forever()
        finally:
            application.close()



# PY3 + PY2
from __future__ import unicode_literals
from __future__ import print_function

from moya.compat import text_type

import re
import unicodedata

_re_html = re.compile(r'<.*?>|\&.*?\;', re.UNICODE | re.DOTALL)
_re_link = re.compile(r'((?:http://|https://|mailto:)[\@\w\.\/]+)', re.UNICODE)
_re_spaceless = re.compile(r'>\s+<')


def escape(text):
    """Escape text for inclusion in html"""
    return text_type(text).replace('&', '&amp;').replace('<', '&lt;').replace('>', '&gt;').replace('"', '&quot;').replace("'", "&#39;")


def slugify(value):
    # Borrowed from Django
    value = unicodedata.normalize('NFKD', value).encode('ascii', 'ignore').decode('ascii')
    value = re.sub('[^\w\s-]', '', value).strip().lower()
    return re.sub('[-\s]+', '-', value)


def textilize(s):
    """Remove markup from html"""
    s = s.replace("<p>", " ").replace('&nbsp;', ' ')
    return _re_html.sub("", s)


def spaceless(s, _re=_re_spaceless):
    """Remove spaces between tags"""
    return _re.sub("><", s)


def summarize(s, max_size=100):
    s = textilize(s)

    if len(s) > max_size:
        if not s[max_size].isspace():
            words = s[:max_size].split()
            words.pop()
        else:
            words = s[:max_size].split()
        s = escape(' '.join(words)) + ' [&hellip;]'
    else:
        s = escape(s)
    return s


def linebreaks(text):
    """Replace new lines with <br>"""
    html = "<br>\n".join(escape(text_type(text)).splitlines())
    return html


def linkify(text):
    def make_link(link):
        if link.startswith('http://') or link.startswith('https://'):
            if link.endswith('.'):
                return '<a href="{}" rel="nofollow">{}</a>.'.format(link[:-1], escape(link[:-1]))
            else:
                return '<a href="{}" rel="nofollow">{}</a>'.format(link, escape(link))
        else:
            email = escape(link.partition(':')[-1])
            if email.endswith('.'):
                return '<a href="mailto:{0}">{0}</a>.'.format(email[:-1])
            else:
                return '<a href="mailto:{0}">{0}</a>'.format(email)

    lines = []
    for line in text.splitlines():
        tokens = []
        pos = 0
        for match in _re_link.finditer(line):
            text = match.group(0)
            start, end = match.span(0)
            if start > pos:
                tokens.append(escape(line[pos:start]))
            tokens.append(make_link(text))
            pos = end
        if pos < len(line):
            tokens.append(escape(line[pos:]))
        lines.append(''.join(tokens))

    return "<br>".join(lines)


if __name__ == "__main__":

    print(spaceless("<b>   <i>   aasd </i>   </b>"))

    print(escape("10 > 5 < 8 & foo"))

    print(summarize('Hello <a href="#">test</a> werwer'))

    print(linkify("My homepage is http://willmcgugan.com/asdd.\nmailto://willmcgugan@gmail.com yadda yadda mailto:willmcgugan@gmail.com "))
    # from moya.tools import timer, MultiReplace

    # replace = MultiReplace({'&': '&amp;', '<' : '&lt;', '>' : '&gt;', '"': '&quote;', "'": "&#39"})

    # text = '''Some "text" & 'stuff'<>'''

    # REPEAT = xrange(100000)


    # with timer('chained escape'):
    #     for _ in REPEAT:
    #         escape(text)

    # with timer('multi escape'):
    #     for _ in REPEAT:
    #         replace(text)



from __future__ import unicode_literals
from __future__ import division

from ..elements.elementbase import LogicElement, Attribute
from ..tags.context import DataSetterBase
from .. import namespaces
from ..compat import implements_to_string

from fs.path import basename, pathjoin, splitext

from PIL import Image, ImageFilter
try:
    from PIL import ExifTags
except ImportError:
    ExifTags = None

import logging
log = logging.getLogger('moya.image')


@implements_to_string
class MoyaImage(object):
    """Proxy interface for a PIL image"""

    def __init__(self, img, filename):
        self._img = img
        self.filename = filename

    def __str__(self):
        w, h = self._img.size
        return "<image '{}' {}x{} {}>".format(self.filename, w, h, self.mode)

    def __repr__(self):
        w, h = self._img.size
        return "<image '{}' {}x{} {}>".format(self.filename, w, h, self.mode)

    def replace(self, img):
        self._img = img

    @property
    def format(self):
        return self._img.format

    @property
    def size(self):
        w, h = self._img.size
        return {'width': w, 'height': h}

    @property
    def mode(self):
        return self._img.mode

    @property
    def info(self):
        return self._img.info

    @property
    def exif(self):
        self._img.load()
        _exif = self._img._getexif()
        if _exif is None:
            return {}

        if ExifTags:
            exif = {
                ExifTags.TAGS[k]: v
                for k, v in self._img._getexif().items()
                if k in ExifTags.TAGS
            }
        else:
            exif = {}

        return exif


class Read(DataSetterBase):
    """Read an image"""
    xmlns = namespaces.image

    class Help:
        synopsis = """read an image from disk"""

    class Meta:
        one_of = [('fsobj', 'fs', 'file')]

    path = Attribute("Path", required=False, default=None)
    fsobj = Attribute("FS", type="Index")
    fs = Attribute("FS name")
    dst = Attribute('Destination', type="reference", default="image")
    _file = Attribute('File object with image data', type="expression")
    filename = Attribute("Filename to associate with the image if the filename can't be detected", default='', type="expression")

    def get_image(self, context, params):
        if self.has_parameter('file'):
            fp = getattr(params.file, '__moyafile__', lambda: params.file)()
            try:
                fp.seek(0)
                img = Image.open(fp)
                img.load()
                fp.seek(0)
            except Exception as e:
                self.throw("image.read-fail", "Failed to read image from file ({})".format(e))
        else:
            if params.fsobj is not None:
                fs = params.fsobj
            else:
                try:
                    fs = self.archive.filesystems[params.fs]
                except KeyError:
                    self.throw("image.no-fs", "No filesystem called '{}'".format(params.fs))
            if params.path is None:
                self.throw("image.path-required", "a path is required when reading from a filesystem")
            try:
                fp = fs.open(params.path, 'rb')
                img = Image.open(fp)
            except Exception as e:
                self.throw("image.read-fail", "failed to read '{}' from {!r} ({})".format(params.path, fs, e))
        return img

    def logic(self, context):
        params = self.get_parameters(context)
        img = self.get_image(context, params)
        try:
            img.load()
        except Exception as e:
            self.throw("image.read-fail", "Failed to read image ({})".format(e))

        moya_image = MoyaImage(img, filename=basename(params.path or params.filename or ''))
        self.set_context(context, params.dst, moya_image)
        log.debug("%r read", moya_image)


class GetSize(Read):
    """Get the dimensions of an image without loading image data, returns a dictionary with keys 'width' and 'height'."""
    xmlns = namespaces.image

    class Help:
        synopsis = """get the dimensions of an image"""

    class Meta:
        one_of = [('fsobj', 'fs', 'file')]

    path = Attribute("Path")
    fsobj = Attribute("FS", type="Index")
    fs = Attribute("FS name")
    dst = Attribute('Destination', type="reference", default="image")
    _file = Attribute('File object with image data', type="expression")

    def logic(self, context):
        params = self.get_parameters(context)
        img = self.get_image(context, params)
        w, h = img.size
        result = {'width': w, 'height': h}
        self.set_context(context, params.dst, result)


class Write(LogicElement):
    """Write an image"""
    xmlns = namespaces.image

    class Help:
        synopsis = """write an image to disk"""

    class Meta:
        one_of = [('fs', 'fsobj')]

    image = Attribute("Image to write", type="expression", default="image", evaldefault=True, missing=False)
    dirpath = Attribute("Directory to write image", required=False, default="/")
    filename = Attribute("Image filename", required=True)
    fsobj = Attribute("FS", type="expression")
    fs = Attribute("FS name")
    format = Attribute("Image format", default=None, choices=['jpeg', 'png', 'gif'])

    def logic(self, context):

        params = self.get_parameters(context)

        if params.fsobj is not None:
            fs = params.fsobj
        else:
            try:
                fs = self.archive.filesystems[params.fs]
            except KeyError:
                self.throw("image.no-fs", "No filesystem called '{}'".format(params.fs))
                return
        path = pathjoin(params.dirpath, params.filename)

        img = params.image._img
        img_format = params.format or splitext(params.filename or '')[-1].lstrip('.') or 'jpeg'

        if img_format == 'jpeg':
            if img.mode != 'RGB':
                img = img.convert('RGB')

        save_params = self.get_let_map(context)
        try:
            with fs.makeopendir(params.dirpath, recursive=True) as dir_fs:
                with dir_fs.open(params.filename, 'wb') as f:
                    img.save(f, img_format, **save_params)
            log.debug("wrote '%s'", params.filename)
        except Exception as e:
            self.throw('image.write-fail', "Failed to write {} to '{}' in {!r} ({})".format(params.image, path, fs, e))


class New(DataSetterBase):
    """Create a blank image."""
    xmlns = namespaces.image

    class Help:
        synopsis = "create a blank image"

    size = Attribute("Size of new image", type="expression", required=True)
    mode = Attribute("Mode", default="RGB")
    color = Attribute("Color", default="#000000")
    dst = Attribute('Destination', type="index", default="image")

    def logic(self, context):
        params = self.get_parameters(context)
        image = Image.new(params.mode, params.size, params.color)
        moya_image = MoyaImage(image, filename="new.jpg")
        self.set_context(context, params.dst, moya_image)


class Copy(DataSetterBase):
    """Create an copy of [c]image[/c] in [c]dst[/c]."""
    xmlns = namespaces.image

    class Help:
        synopsis = "create a copy of an image"

    image = Attribute("Image to copy", type="expression", default="image", evaldefault=True)
    dst = Attribute('Destination', type="reference", default="image")

    def logic(self, context):
        params = self.get_parameters(context)
        img = params.image._img
        moya_image = MoyaImage(img.copy(), params.image.filename)
        self.set_context(context, params.dst, moya_image)


class Show(LogicElement):
    """Show an image (for debugging purposes). Imagemagick is required for this operation."""
    xmlns = namespaces.image

    class Help:
        synopsis = "show an image"

    image = Attribute("Image to show", type="expression", default="image", evaldefault=True)

    def logic(self, context):
        if context['.debug']:
            self.image(context)._img.show()
        else:
            log.warn('<show> can be used in debug mode only')


_resample_methods = {
    "nearest": Image.NEAREST,
    "bilinear": Image.BILINEAR,
    "bicubic": Image.BICUBIC,
    "antialias": Image.ANTIALIAS
}


def _fit_dimensions(image, width, height):
    w, h = image.size
    if width is None:
        width = w * (height / h)
    if height is None:
        height = h * (width / w)

    ratio = min(width / w, height / h)
    width = w * ratio
    height = h * ratio

    return int(round(width)), int(round(height))


class ResizeToFit(LogicElement):
    """Resize image to fit within the given dimensions (maintains aspect ratio)."""

    xmlns = namespaces.image

    class Help:
        synopsis = "resize an image to fit within new dimensions"

    image = Attribute("Image to show", type="expression", default="image", evaldefault=True)
    width = Attribute("New width", type="integer", required=False, default=None,)
    height = Attribute("New height", type="integer", required=False, default=None)
    resample = Attribute("Method for resampling", default="antialias", choices=_resample_methods.keys())

    def logic(self, context):
        params = self.get_parameters(context)
        image = params.image._img
        new_size = _fit_dimensions(image, params.width, params.height)
        w, h = new_size
        if not w or not h:
            self.throw('image.bad-dimensions',
                       'Invalid image dimensions ({} x {})'.format(params.width, params.height),
                       diagnosis="Width and / or height should be supplied, and should be non-zero")
        params.image.replace(image.resize(new_size, _resample_methods[params.resample]))


class ZoomToFit(LogicElement):
    """Resize image to given dimensions, cropping if necessary."""

    xmlns = namespaces.image

    class Help:
        synopsis = "resize an image to fit in new dimensions, with cropping"

    image = Attribute("Image to show", type="expression", default="image", evaldefault=True)
    width = Attribute("New width", type="integer", required=True)
    height = Attribute("New height", type="integer", required=True)
    resample = Attribute("Method for resampling", default="antialias", choices=_resample_methods.keys())

    def logic(self, context):
        params = self.get_parameters(context)
        image = params.image._img

        aspect = image.size[0] / image.size[1]
        if image.size[0] > image.size[1]:
            new_size = int(params.height * aspect), params.height
        else:
            new_size = params.width, int(params.width / aspect)

        img = image.resize(new_size, _resample_methods[params.resample])

        w = params.width
        h = params.height
        x = (img.size[0] - w) // 2
        y = (img.size[1] - h) // 2

        box = x, y, x + w, y + h
        params.image.replace(img.crop(box))


class Resize(LogicElement):
    """Resize an image to new dimensions."""
    xmlns = namespaces.image

    class Help:
        synopsis = "resize an image"

    image = Attribute("Image to show", type="expression", default="image", evaldefault=True)
    width = Attribute("New width", type="integer", required=False, default=None,)
    height = Attribute("New height", type="integer", required=False, default=None)
    resample = Attribute("Method for resampling", default="antialias", choices=["nearest", "bilinear", "bicubic", "antialias"])

    def logic(self, context):
        params = self.get_parameters(context)
        image = params.image._img
        new_size = (params.width, params.height)
        w, h = new_size
        if not w or not h:
            self.throw('image.bad-dimensions',
                       'Invalid image dimensions ({} x {})'.format(params.width, params.height),
                       diagnosis="Width and / or height should be supplied, and should be non-zero")
        params.image.replace(image.resize(new_size, _resample_methods[params.resample]))


class ResizeCanvas(LogicElement):
    """Resize the image canvas."""

    xmlns = namespaces.image

    class Help:
        synopsis = "resize the image canvas"

    image = Attribute("Image to show", type="expression", default="image", evaldefault=True)
    width = Attribute("New width", type="integer", required=True)
    height = Attribute("New height", type="integer", required=True)
    color = Attribute("Background color", type="color", default="#000000")

    def logic(self, context):
        image, w, h, color = self.get_parameters(context, 'image', 'width', 'height', 'color')
        img = image._img
        mode = img.mode
        if color.a != 1:
            mode = 'RGBA'
        new_img = Image.new(mode, (w, h), color.as_pillow_tuple())

        iw, ih = img.size

        x = (w - iw) // 2
        y = (h - ih) // 2

        new_img.paste(img, (x, y))
        image.replace(new_img)


class Square(LogicElement):
    """Square crop an image."""

    xmlns = namespaces.image

    class Help:
        synopsis = "square crop an image"

    image = Attribute("Image to crop", type="expression", default="image", evaldefault=True)

    def logic(self, context):
        params = self.get_parameters(context)
        img = params.image._img

        w, h = img.size
        size = min(img.size)
        x = (w - size) // 2
        y = (h - size) // 2

        new_img = Image.new(img.mode, (size, size))
        new_img.paste(img, (-x, -y))

        params.image.replace(new_img)


class Crop(LogicElement):
    """Crop an image to a given area."""
    xmlns = namespaces.image

    class Help:
        synopsis = "crop an image"

    image = Attribute("Image to show", type="expression", default="image", evaldefault=True)
    box = Attribute("Crop size (either [width, height] or [x, y, width, height])", type="expression")

    def logic(self, context):
        params = self.get_parameters(context)
        size = params.box

        img = params.image._img
        if len(size) == 2:
            w, h = size
            x = (img.size[0] - w) // 2
            y = (img.size[1] - h) // 2
        elif len(size) == 4:
            x, y, w, h = size
        else:
            self.throw('bad-value.box-invalid',
                       "parameter 'box' should be  sequence of 2 or 4 parameters (not {})".format(context.to_expr(size)))

        box = x, y, x + w, y + h
        params.image.replace(img.crop(box))


class GaussianBlur(LogicElement):
    """Guasian blur an image"""
    xmlns = namespaces.image

    image = Attribute("Image to show", type="expression", default="image", evaldefault=True)
    radius = Attribute("Radius of blur", type="integer", default=2)

    def logic(self, context):
        params = self.get_parameters(context)
        img = params.image._img
        new_image = img.filter(ImageFilter.GaussianBlur(radius=params.radius))
        params.image.replace(new_image)



from __future__ import unicode_literals


from ..tags.context import DataSetter
from ..elements import Attribute
from .. import errors
from moya.compat import text_type


class ElementContainer(object):

    def __init__(self, archive):
        self.archive = archive

    def __getitem__(self, elementid):
        return self.archive.get_element(elementid)

    def __contains__(self, elementid):
        try:
            self[elementid]
        except errors.ElementNotFoundError:
            return False
        else:
            return True


class Inspect(DataSetter):
    """
    Get an object which contains all the elements in the project.
    """

    class Help:
        synopsis = "inspect elements in the project"

    def get_value(self, context):
        return ElementContainer(self.archive)


class QualifyElementref(DataSetter):
    """
    Qualifies an element reference.

    """

    app = Attribute("An application name", type="expression", default=None)
    lib = Attribute("A library name", type="expression", default=None)
    ref = Attribute("An element reference", type="expression", required="yes")

    class Meta:
        one_of = [('app', 'lib')]

    class Help:
        synopsis = "create an element reference with an app/lib name"

    def get_value(self, context):
        app, lib, ref = self.get_parameters(context, 'app', 'lib', 'ref')
        if app is not None:
            app = self.archive.get_app(app)
        if lib is not None:
            lib = self.archive.get_lib(lib)
        ref = text_type(ref)
        try:
            element = self.archive.get_element(ref, app=app, lib=lib or None)
        except Exception as e:
            self.throw('qualify-elementref.not-found',
                       'unable to look up element ({})'.format(e))
        element_ref = "{}#{}".format(element.app.name, element.element.libname)
        return element_ref



"""A container for Moya code tracebacks"""

from __future__ import print_function

from .console import Console, Cell
from .template.errors import TemplateError
from .context.expression import ExpressionError
from .context.errors import SubstitutionError
from .moyaexceptions import MoyaException
from .compat import implements_to_string, text_type
from .traceframe import Frame

import io
import sys
from operator import attrgetter
import traceback as pytraceback


_PYTHON_ERROR_TEXT = """A Python Exception may indicate either a bug in a Python extension, or Moya itself.

Consider reporting this to the Moya developers."""


@implements_to_string
class Traceback(object):

    _get_frame_compare = attrgetter('_location', 'lineno', 'libid')

    def __init__(self, url=None, method=None, handler=None, exc=None):
        self.url = url
        self.method = method
        self.handler = handler
        self.stack = []
        self.exception = None
        self.tb = None
        self.error_message = None
        self.exc = exc
        self.exc_info = None
        self.msg = None
        self.error_type = "internal error"
        self._displayed = False
        self.diagnosis = getattr('exc', 'diagnosis', None)

    @property
    def console_error(self):
        if hasattr(self.exc, '__moyaconsole__') and getattr(self.exc.__moyaconsole__, 'is_default_error_message', False):
            return None
        console = Console(html=True)
        console.obj(None, self.exc)
        return console.get_text()

    def remove_duplicates(self):
        current = None
        out = []
        for frame in self.stack:
            if current is None or self._get_frame_compare(frame) != self._get_frame_compare(current):
                out.append(frame)
                current = frame
        self.stack = out

    def add_frame(self, frame):
        self.stack.append(frame)

    def __str__(self):
        console = Console(text=True)
        self.__moyaconsole__(console)
        return console.get_text()

    def __moyaconsole__(self, console):
        stack = self.stack
        console.div("Logic Error", bold=True, fg="red")
        for frame in stack:
            console.wraptext(frame.location)
            if frame.one_line:
                console.pysnippet("\n" * (frame.lineno - 1) + frame.code, frame.lineno, extralines=0)
            elif frame.code:
                if frame.format == 'xml':
                    console.xmlsnippet(frame.code, frame.lineno, extralines=3)
                elif frame.format == 'moyatemplate':
                    start, end = frame.cols
                    console.templatesnippet(frame.code,
                                            lineno=frame.lineno,
                                            colno=start,
                                            endcolno=end)
                else:
                    console.pysnippet(frame.code, frame.lineno, extralines=3)
            console.nl()

        if self.tb:
            console.exception(self.tb, tb=True)
        else:
            console.error(self.msg)
        if self.diagnosis:
            console.table([[Cell(self.diagnosis, italic=True)]])
        console.div()


def build(context, stack, node, exc, exc_info, request, py_traceback=True):

    add_pytraceback = True
    if node is not None:
        node = getattr(node, 'node', node)
    if stack is None:
        stack = context.get('._callstack', [])

    if request is not None:
        traceback = Traceback(request.path_info, request.method, exc=exc)
    else:
        traceback = Traceback(exc=exc)
    traceback.diagnosis = getattr(exc, 'diagnosis', None)

    add_pytraceback = not getattr(exc, 'hide_py_traceback', False)
    traceback.error_type = getattr(exc, 'error_type', 'internal error')

    base = context.get('.sys.base', '')

    def relativefrom(base, path):
        if base and path.startswith(base):
            path = "./" + path[len(base):]
        return path

    for s in stack:
        e = getattr(s, 'element', None)
        if e and e._code:
            frame = Frame(e._code,
                          e._location,
                          e.source_line or 1,
                          obj=text_type(e),
                          libid=e.libid)
            traceback.add_frame(frame)

    element = getattr(exc, 'element', None)

    if element is not None and hasattr(element.document, 'structure'):
        frame = Frame(element.document.structure.xml,
                      element._location,
                      element.source_line or 1,
                      obj=text_type(element),
                      libid=element.libid)
        traceback.add_frame(frame)
        add_pytraceback = False

    elif hasattr(node, '_location') and hasattr(node, 'source_line'):
        if node._code:
            frame = Frame(node._code,
                          node._location,
                          node.source_line or 1,
                          obj=text_type(node),
                          libid=node.libid)
            traceback.add_frame(frame)

    if isinstance(exc, MoyaException):
        traceback.error_type = "Moya Exception"
        traceback.moya_exception_type = exc.type
        add_pytraceback = False

    elif isinstance(exc, ExpressionError):
        traceback.error_type = "Expression Error"
        add_pytraceback = False

    elif isinstance(exc, SubstitutionError):
        traceback.error_type = "Substitution Error"
        add_pytraceback = False

    elif isinstance(exc, TemplateError):
        traceback.error_type = "Template Error"

    traceback.exception = exc
    traceback.msg = text_type(exc)
    traceback.diagnosis = traceback.diagnosis or getattr(exc, 'diagnosis', None)

    if hasattr(exc, 'get_moya_frames'):
        mf = exc.get_moya_frames()
        traceback.stack.extend(mf)

    if context.get('.develop', False):
        add_pytraceback = True

    if add_pytraceback and exc_info and py_traceback:
        traceback.error_type = "Python Exception"
        tb_type, tb_value, tb = exc_info
        traceback.tb = ''.join(pytraceback.format_exception(tb_type, tb_value, tb))

        pyframes = pytraceback.extract_tb(tb)

        for i, f in enumerate(reversed(pyframes)):
            if f[2] == 'logic':
                pyframes = pyframes[len(pyframes) - i - 1:]
                break

        for (filename, line_number, function_name, text) in pyframes:
            one_line = False
            try:
                with io.open(filename, 'rt') as f:
                    code = f.read()
            except:
                code = text
                one_line = True

            code_path = relativefrom(base, filename)
            frame = Frame(code,
                          code_path,
                          line_number,
                          one_line=one_line,
                          obj=function_name,
                          format="python")
            traceback.add_frame(frame)
            traceback.msg = text_type(exc)
        if traceback.diagnosis is None:
            traceback.diagnosis = _PYTHON_ERROR_TEXT

    traceback.remove_duplicates()

    return traceback


def format_trace(context, stack, node, exc_info=None):
    if exc_info is None:
        exc_info = sys.exc_info()
    request = context.get('.request', None)
    moya_trace = build(context, stack, None, node, exc_info, request, py_traceback=False)
    return text_type(moya_trace)



from __future__ import unicode_literals
from __future__ import print_function

from ...command import SubCommand

import sys
import io


class Syntax(SubCommand):
    """
    A tool to syntax highlight a file.

    Moya uses this in the debug library to syntax highlight files and errors.

    The syntax highlighter is somewhat naive. There maybe better general purpose syntax highlighters out there...

    """
    help = """syntax highlight a file"""

    def add_arguments(self, parser):
        parser.add_argument(dest='path', metavar="PATH",
                            help="Path to a file to highlight")
        parser.add_argument('-f', '--format', metavar="format", default='text',
                            help="format to highlight")
        parser.add_argument('-l', '--lines', dest="lines", action="store_true", default=False,
                            help="render lines numbers")
        return parser

    def run(self):
        args = self.args
        from moya import syntax
        with io.open(args.path, 'rt') as f:
            code = f.read()
        html = syntax.highlight(args.format, code, line_numbers=args.lines)
        sys.stdout.write(html + '\n')



from ...command import SubCommand
from ...wsgi import WSGIApplication
from ...console import ConsoleHighlighter
from ...tools import url_join
from ...http import StatusCode

class URLHighlight(ConsoleHighlighter):
    styles = {
        None: "bold blue",
        "asterix": "bold yellow",
        "param": "bold magenta"
    }

    highlights = [
        r'(?P<param>\{.*?\})',
        r'(?P<asterix>\*)',
    ]


class TargetHighlight(ConsoleHighlighter):
    styles = {
        "lib": "bold blue",
        "name": "bold green",
        "hash": "white"
    }

    highlights = [
        r'(?P<lib>.*?)(?P<hash>#)(?P<name>.*?)[,\s$]',
    ]


class URLS(SubCommand):
    """List URLs"""
    help = "list URLs"

    def add_arguments(self, parser):
        parser.add_argument("-l", "--location", dest="location", default=None, metavar="PATH",
                            help="location of the Moya server code")
        parser.add_argument("-i", "--ini", dest="settings", default=None, metavar="SETTINGSPATH",
                            help="path to project settings file")
        parser.add_argument("--server", dest="server", default="main", metavar="SERVERREF",
                            help="server element to use")
        parser.add_argument("--sort", "-s", dest="sort", default=False, action="store_true",
                            help="sort URL output")
        parser.add_argument('-m', '--method', dest="method", default=None, metavar="VERB",
                            help="only show urls for the given method")
        parser.add_argument('-a', '--app', dest="app", metavar="APPLICATION", default=None,
                            help="only show urls for given app")
        parser.add_argument('--handler', default=None, metavar="STATUS",
                            help="show handlers for the given error")

    def run(self):
        args = self.args

        application = WSGIApplication(self.location,
                                      self.get_settings(),
                                      args.server,
                                      disable_autoreload=True)
        urlmapper = application.server.urlmapper
        urls = []

        def render_urlmapper(app_name, parent_url, urlmapper):
            for route in urlmapper.routes:
                if route.partial:
                    render_urlmapper(app_name, url_join(parent_url, route.route), route.target)
                else:
                    if args.handler:
                        handler = StatusCode(args.handler)
                        if handler not in route.handlers:
                            continue
                    else:
                        if args.method and not route.match_method(args.method, None):
                            continue
                        if route.handlers:
                            continue
                        if args.app and args.app != app_name:
                            continue
                    urls.append((
                                ", ".join(route.methods or ['*']),
                                url_join(parent_url, route.route),
                                app_name,
                                route.name or '',
                                ", ".join(route.target)))

        for route in urlmapper.routes:
            app_name = route.name or ''
            render_urlmapper(app_name, route.route, route.target)

        if args.sort:
            urls.sort(key=lambda row: row[1])

        def highlight_url(text):
            return URLHighlight.highlight(text)

        def highlight_target(text):
            return TargetHighlight.highlight(text)

        self.console.table(urls,
                           ["method(s)", "URL", "app", "name", "target(s)"],
                           cell_processors={1: highlight_url, 4: highlight_target})



from __future__ import unicode_literals
from __future__ import absolute_import

import sys
from cmd import Cmd
try:
    import readline
    readline  # hides a warning
except ImportError:
    pass

from .console import Cell
from .elements.help import help
from . import namespaces
from .tools import extract_namespace
from .compat import text_type


def winpdb():
    import rpdb2
    rpdb2.start_embedded_debugger("password")


# Global containing last command, None will show help
_previous_command = None


class MoyaCmdDebugger(Cmd):
    prompt = "moya>"

    command_help = [
        (
            ' let foo="bar"',
            """
            Set foo to \"bar\" in the context.
            """
        ),
        (
            ' foo',
            """
            Evaulate the 'foo' in the current context and display the result. Any valid expression will also work, for example 1 + 1, "Hello," + "World!" etc.
            """
        ),
        (
            ' foo?',
            """
            Convert the value of 'foo' to text.

            """
        ),
        (
            ' foo??',
            """
            Convert the value of 'foo' to a Moya expression (if possible).

            """
        ),
        (
            ' foo???',
            """
            Covert the value of 'foo' to the internal (Python) representation.
            """
        ),
        (
            ' $$',
            """
            Show the current scope
            """
        ),
        (
            't, stack [EXTRALINES]',
            """
            Show the current call stack. If EXTRALINES is provided, it should be an integer indicating the number of lines of code to show in the stack trace.
            """
        ),
        (
            "s, step",
            """
            Advance to next logic element.
            """
        ),
        (
            "o, over",
            """
            Step over the next logic element.
            """
        ),
        (
            'u, out',
            """
            Step out of the current call.
            """
        ),
        (
            "c, continue",
            """
            Run until the next breakpoint, or to the end of the logic code.

            """
        ),
        (
            "help",
            """
            Show this help information.
            """
        ),
        (
            "help TAG",
            """
            Show help information for a given tag.
            """
        ),
        (
            "w, where [EXTRALINES]",
            """
            Show the current position in moya code, if EXTRALINES it provided, it should be an integer indicating the number of additional lines either side of the current position to display.
            """
        ),
        (
            "watch",
            """
            When followed by an expression, it will be added to the watch list (a table of expressions and their results). When given without an argument, the watch list will be reset.
            """
        ),
        (
            "winpdb PASSWORD",
            """
            Launch WinPDB to debug the next Python call.
            """
        ),
        (
            "ctrl+D",
            """
            Exit debugger and continue with moya code execution. Ignores all further breakpoints to the end of the request.

            """
        ),
        (
            "<ENTER>",
            """
            Repeat last command.
            """
        ),
        (
            "v, view",
            """
            Display the full view of the currently executing moya code.

            """
        ),
        (
            "e, eval",
            """
            Evaluates an expression, or shows the current frame if no argument is given.
            """
        ),
        (
            "exit",
            "Stop execution of Moya code and exit debugger."
        ),
        (
            "r, run",
            "Exit debugger and continue with moya code execution. Ignores all further breakpoints in the session."
        )
    ]

    def __init__(self, archive, console):
        self.console = console
        self.archive = archive
        Cmd.__init__(self)

    def default(self, line):
        if not isinstance(line, text_type):
            line = line.decode(sys.stdin.encoding)
        self.usercmd = line
        global _previous_command
        _previous_command = line

    def postcmd(self, stop, line):
        if stop is False:
            return False
        if _previous_command is None:
            return self.do_help()
        self.usercmd = _previous_command
        return True

    def do_help(self, line=''):
        if line:
            ns, tag = extract_namespace(line)
            self.console.div("Help on <%s/>" % tag, fg="blue", bold=True)
            help(self.archive, self.console, "{%s}%s" % (ns or namespaces.default, tag))
            return False

        self.console.div("Moya Debugger", fg="blue", bold=True)
        self.console.text("Moya's Debugger http://www.moyaproject.com/debugger/", fg="black", bold=True).nl()
        table = [(Cell("Command", bold=True), Cell("Description", bold=True))]
        command_help = sorted(self.command_help, key=lambda h: h[0])

        def format_desc(desc):
            lines = [l.strip() for l in desc.splitlines() if l.strip()]
            return "\n".join(lines)
        for i, (command, desc) in enumerate(command_help):
            table.append([Cell(command, bold=True, fg="green"),
                          Cell(format_desc(desc))])
        self.console.table(table, header=True)
        return False



from __future__ import unicode_literals

__author__ = "Will McGugan <admin@moyaproject.com>"
__version__ = "0.5.28"
# *** Don't forget to update version in setup.py ***

__all__ = ['pilot', 'expose']


from . import expose
from .moyapilot import Pilot
from .compat import text_type, PY2


pilot = Pilot()

import decimal

if PY2:

    decimal.Decimal.__moyarepr__ = lambda self, context: "d:'{}'".format(text_type(self))
    _decimal_str = decimal.Decimal.__str__
    decimal.Decimal.__str__ = lambda self: "{:f}".format(self.normalize())
    decimal.Decimal.__unicode__ = decimal.Decimal.__str__


else:
    _decimal_normalize = decimal.Decimal.normalize

    class MoyaDecimal(decimal.Decimal):

        def __moyarepr__(self, context):
            return "d:'{}'".format(_decimal_normalize(self))

        def __str__(self):
            return "{:f}".format(_decimal_normalize(self))

        def __unicode__(self):
            return "{:f}".format(_decimal_normalize(self))

    decimal.Decimal = MoyaDecimal



from __future__ import unicode_literals
from __future__ import print_function

from ..context.tools import to_expression
from ..compat import (number_types,
                      implements_to_string,
                      xrange,
                      implements_bool,
                      string_types,
                      unichr)

__all__ = ["ExpressionRange",
           "ExclusiveIntegerRange",
           "InclusiveIntegerRange",
           "ExclusiveCharacterRange",
           "InclusiveCharacterRange"]


@implements_bool
@implements_to_string
class ExpressionRange(object):
    inclusive = True

    @classmethod
    def create(cls, context, start, end, inclusive=True):
        if not isinstance(start, number_types + string_types):
            if hasattr(start, '__moyarange__'):
                moyarange = start.__moyarange__(context, end, inclusive=inclusive)
            else:
                raise ValueError("Can't create range from {!r} to {!r}".format(start, end))
            return moyarange
        if isinstance(start, number_types):
            if inclusive:
                return InclusiveIntegerRange(context, start, end)
            else:
                return ExclusiveIntegerRange(context, start, end)
        elif isinstance(start, string_types):
            if inclusive:
                return InclusiveCharacterRange(context, start, end)
            else:
                return ExclusiveCharacterRange(context, start, end)
        raise ValueError("Can't create range from {} to {}".format(to_expression(context, start),
                                                                   to_expression(context, end)))

    def __init__(self, context, start, end):
        self.start = start
        self.end = end
        try:
            self._build(start, end)
        except TypeError:
            raise ValueError("Can't create range between {} and {}".format(to_expression(context, start),
                                                                           to_expression(context, end)))

    def __str__(self):
        return "<range {} to {} inclusive>".format(self.start, self.end)

    def _build(self, start, end):
        raise NotImplementedError

    # def __moyaconsole__(self, console):
    #     return console.text(unicode(self))

    def __moyarepr__(self, context):
        if self.inclusive:
            return "({})..({})".format(to_expression(context, self.start),
                                       to_expression(context, self.end))
        else:
            return "({})...({})".format(to_expression(context, self.start),
                                        to_expression(context, self.end))

    def keys(self):
        return range(0, len(self))

    def values(self):
        return list(self)

    def items(self):
        return list(enumerate(self))

    def __bool__(self):
        return bool(len(self))

    def __add__(self, value):
        return list(self) + [value]

    def __radd__(self, value):
        return [value] + list(self)

    def __sub__(self, value):
        l = list(self)
        l.remove(value)
        return l

    def __getitem__(self, index):
        try:
            index = int(index)
        except TypeError:
            raise
        for i, v in enumerate(self):
            if i == index:
                return v
        raise KeyError(index)

    def __contains__(self, value):
        return any(v == value for v in self)

    def __len__(self):
        return sum(1 for _ in self)


@implements_to_string
class ExclusiveIntegerRange(ExpressionRange):
    inclusive = False

    def _build(self, start, end):
        self._a = int(start)
        self._b = int(end)
        self._forward = end >= start

    def __str__(self):
        return "<range {} to {} exclusive>".format(self.start, self.end)

    def __iter__(self):
        if self._forward:
            return iter(xrange(self._a, self._b))
        else:
            return iter(xrange(self._a, self._b, -1))

    def __contains__(self, v):
        try:
            v = int(v)
        except TypeError:
            return False
        if self._forward:
            return v >= self._a and v < self._b
        else:
            return v <= self._a and v > self._b

    def __len__(self):
        return abs(self._a - self._b)


@implements_to_string
class InclusiveIntegerRange(ExpressionRange):
    inclusive = True

    def _build(self, start, end):
        self._a = int(start)
        self._b = int(end)
        self._forward = end >= start

    def __str__(self):
        return "<range {} to {} inclusive>".format(self.start, self.end)

    def __iter__(self):
        if self._forward:
            return iter(xrange(self._a, self._b + 1))
        else:
            return iter(xrange(self._a, self._b - 1, -1))

    def __contains__(self, v):
        try:
            v = int(v)
        except TypeError:
            return False
        if self._forward:
            return v >= self._a and v <= self._b
        else:
            return v <= self._a and v >= self._b

    def __len__(self):
        return abs(self._a - self._b) + 1


@implements_to_string
class ExclusiveCharacterRange(ExpressionRange):
    inclusive = False

    def _build(self, start, end):
        self._a = ord(start[0])
        self._b = ord(end[0])
        self._forward = end >= start

    def __str__(self):
        return "<range {} to {} exclusive>".format(self.start, self.end)

    def __iter__(self):
        if self._forward:
            return iter(unichr(c) for c in xrange(self._a, self._b))
        else:
            return iter(unichr(c) for c in xrange(self._a, self._b, -1))

    def __contains__(self, v):
        if not isinstance(v, string_types):
            return False
        if len(v) != 1:
            return False
        v = ord(v[0])
        if self._forward:
            return v >= self._a and v < self._b
        else:
            return v <= self._a and v > self._b

    def __len__(self):
        return abs(self._a - self._b)


@implements_to_string
class InclusiveCharacterRange(ExpressionRange):
    invlusive = True

    def _build(self, start, end):
        self._a = ord(start[0])
        self._b = ord(end[0])
        self._forward = end >= start

    def __str__(self):
        return "<range {} to {} inclusive>".format(self.start, self.end)

    def __iter__(self):
        if self._forward:
            return iter(unichr(c) for c in xrange(self._a, self._b + 1))
        else:
            return iter(unichr(c) for c in xrange(self._a, self._b - 1, -1))

    def __contains__(self, v):
        if not isinstance(v, string_types):
            return False
        if len(v) != 1:
            return False
        v = ord(v[0])
        if self._forward:
            return v >= self._a and v <= self._b
        else:
            return v >= self._b and v <= self._a

    def __len__(self):
        return abs(self._a - self._b) + 1


if __name__ == "__main__":
    r = ExpressionRange.create(0.0, 10, inclusive=False)
    print(len(r))
    print(list(r))
